# -*- coding: utf-8 -*-
"""
As BioSTEAM objects are created, they are automatically registered. The `find` object allows the user to find any Unit, Stream or System instance.  When `find` is called, it simply looks up the item and returns it. 

"""
from thermosteam.utils import Registry
from ._digraph import make_digraph, save_digraph
from thermosteam import Stream
from . import Unit, System

__all__ = ('find', 'Flowsheet')

# %% Flowsheet search      

class Flowsheets:
    __getattribute__ = __getitem__ = object.__getattribute__
    
    def __setattr__(self, key, value):
        raise TypeError(f"'{type(self).__name__}' object does not support attribute assignment")
    __setitem__ = __setattr__
    
    def __iter__(self):
        yield from self.__dict__.values()
    
    def __delattr__(self, key):
        if key == find.ID:
            raise AttributeError('cannot delete main flowsheet')
        else:
            super().__delattr__(key)
    
    def __repr__(self):
        return f'Register:\n ' + '\n '.join([repr(i) for i in self])
    
    
class Flowsheet:
    """Create a Flowsheet object which stores references to all stream, unit, and system objects."""
    
    #: [Register] All flowsheets.
    flowsheet = Flowsheets()
    
    def __init__(self, ID):        
        #: [Register] Contains all System objects as attributes.
        self.system = Registry()
        
        #: [Register] Contains all Unit objects as attributes.
        self.unit = Registry()
        
        #: [Register] Contains all Stream objects as attributes.
        self.stream = Registry()
        
        #: [str] ID of flowsheet.
        self._ID = ID
        self.flowsheet.__dict__[ID] = self
    
    def __setattr__(self, key, value):
        if hasattr(self, '_ID'):
            raise TypeError(f"'{type(self).__name__}' object does not support attribute assignment")
        else:
            super().__setattr__(key, value)
    
    @property
    def ID(self):
        return self._ID
    
    def diagram(self, kind='surface', file=None, format='svg', **graph_attrs):
        """Display all units and attached streams.
        
        Parameters
        ----------
        kind='surface' : Must be one of the following
            * **'thorough':** Thoroughly display every unit.
            * **'surface':** Display units and recycle systems.
            * **'minimal':** Minimally display flowsheet as a box with feeds and products.
        
        """
        if kind == 'thorough':
            return self._thorough_diagram(file, format, **graph_attrs)
        elif kind == 'surface':
            return self._surface_diagram(file, format, **graph_attrs)
        elif kind == 'minimal':
            return self._minimal_diagram(file, format, **graph_attrs)
        else:
            raise ValueError(f"kind must be either 'thorough', 'surface', or 'minimal'.")
    
    def _thorough_diagram(self, file, format, **graph_attrs):
        units = list(self.unit)
        units.reverse()
        streams = set()
        for u in units:
            streams.update(u._ins)
            streams.update(u._outs)
        f = make_digraph(units, streams, format=format, **graph_attrs)
        save_digraph(f, file, format)
    
    def _minimal_diagram(self, file, format, **graph_attrs):
        from . import _system
        streams = list(self.stream)
        feeds = set(filter(_system._isfeed, streams))
        products = set(filter(_system._isproduct, streams))
        product = Stream(None)
        product._ID = ''
        feed = Stream(None)
        feed._ID = ''
        _system._streamUnit('\n'.join([i.ID for i in feeds]),
                           None, feed)
        _system._streamUnit('\n'.join([i.ID for i in products]),
                           product, None)
        unit = _system._systemUnit(self.ID, feed, product)
        unit.line = 'flowsheet'
        unit.diagram(1, file, format, **graph_attrs)
        
    def _surface_diagram(self, file, format, **graph_attrs):
        from . import _system
        units = set(self.unit)
        StrUnit = _system._streamUnit
        refresh_units = set()
        for i in self.system:
            if i.recycle and not any(sub.recycle for sub in i.subsystems):
                outs = []
                ins = []
                feeds = []
                products = []
                for s in i.streams:
                    source = s._source
                    sink = s._sink
                    if source in i.units and sink not in i.units:
                        if sink: outs.append(s)
                        else: products.append(s)
                        refresh_units.add(source)
                    elif sink in i.units and source not in i.units:
                        if source: ins.append(s)
                        else: feeds.append(s)
                        refresh_units.add(sink)
                
                if len(feeds) > 1:
                    feed = Stream(None)
                    feed._ID = ''
                    units.add(StrUnit('\n'.join([i.ID for i in feeds]),
                                      None, feed))
                    ins.append(feed)
                else: ins += feeds
                
                if len(products) > 1:
                    product = Stream(None)
                    product._ID = ''
                    units.add(StrUnit('\n'.join([i.ID for i in products]),
                                      product, None))
                    outs.append(product)
                else: outs += products
                
                subsystem_unit = _system._systemUnit(i.ID, ins, outs)
                units.difference_update(i.units)
                units.add(subsystem_unit)
        
        sys = _system.System(None, units)
        sys._thorough_diagram(file, format, **graph_attrs)
        for i in refresh_units:
            i._ins[:] = i._ins
            i._outs[:] = i._outs
    
    def __call__(self, ID):
        """Return requested biosteam item.
    
        Parameters
        ----------
        ID : str
              ID of the requested item.
    
        """
        ID = ID.replace(' ', '_')
        obj = (self.stream.search(ID)
               or self.unit.search(ID)
               or self.system.search(ID))
        if not obj: raise LookupError(f"no registered item '{ID}'")
        return obj
    
    def __str__(self):
        return self.ID
    
    def __repr__(self):
        return f'<{type(self).__name__}: {self.ID}>'


class MainFlowsheet(Flowsheet):
    """Create a MainFlowsheet object which automatically registers biosteam objects as they are created."""
    __slots__ = ()
    
    @staticmethod
    def set_flowsheet(flowsheet):
        """Set main flowsheet that is updated with new biosteam objects."""
        if isinstance(flowsheet, Flowsheet):
            dct = flowsheet.__dict__
        elif isinstance(flowsheet, str):
            if flowsheet in find.flowsheet:
                dct = find.flowsheet[flowsheet].__dict__
            else:
                new_flowsheet = Flowsheet(flowsheet)
                find.flowsheet.__dict__[flowsheet] = new_flowsheet
                dct = new_flowsheet.__dict__
        else:
            raise TypeError('flowsheet must be a Flowsheet object')
        Stream.registry = dct['stream']
        System.registry = dct['system']
        Unit.registry = dct['unit']
        object.__setattr__(find, '__dict__', dct)
        
    __setattr__ = Flowsheets.__setattr__
    
    def __new__(cls):
        raise TypeError('cannot create new find object.')

    def __repr__(self):
        return f'<{type(self).__name__}: {self.ID}>'
    
    
#: [find] Find BioSTEAM objects by ID.
find = object.__new__(MainFlowsheet)
find.set_flowsheet('default')
    

# %% Attempt at contantly rendered digraph

# #: [bool] True if flowsheet is contantly rendered
# self.live = True

# def _show_once(self):
    #     fig = plt.figure()
    #     plt.axis('off')
    #     f = open('diagram.png', 'wb')
    #     diagram = make_digraph(self.unit.values()).pipe(format='png')
    #     f.write(diagram)
    #     img = mpimg.imread('diagram.png')
    #     plt.imshow(img, interpolation='spline36')
    #     fig.show()
    #     # img = Image.open('diagram.png')
    #     # img.show() 
    
    # def _show(self):
    #     fig = plt.figure()
    #     plt.axis('off')
    #     f = open('diagram.png', 'wb')
    #     for i in range(3):
    #         diagram = make_digraph(self.unit.values()).pipe(format='png')
    #         f.write(diagram)
    #         img = mpimg.imread('diagram.png')
    #         plt.clf() 
    #         plt.imshow(img, interpolation='spline36')
    #         fig.show()
    #         time.sleep(10)
    #     f.close()
    #     os.remove('diagram.png')
    
    # def show(self):
    #     t1 = threading.Thread(target=self._show())
    #     t1.start()


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 23 20:03:24 2020

@author: yoelr
"""
import numpy as np
import multiprocessing as mp
import pandas as pd

__all__ = ('evaluate_coordinate_in_parallel',)

def evaluate_coordinate_in_parallel(f_evaluate_at_coordinate, coordinate,
                                    metrics, multi_coordinate=False,
                                    name=None, names=None, xlfile=None): 
    # In parallel
    pool = mp.Pool(mp.cpu_count())
    map = pool.starmap if multi_coordinate else pool.map
    tables = map(f_evaluate_at_coordinate, coordinate)
    pool.close()
    
    # Initialize data containers
    table = tables[0]
    N_samples, N_parameters = table.shape
    N_coordinate = len(coordinate)
    metric_data = {i.index: np.zeros([N_samples, N_coordinate])
                   for i in metrics}
    try:
        for n, table in enumerate(tables):
            for metric in metric_data:
                metric_data[metric][:, n] = table[metric]
    except:
        return tables, metric_data
    
    if xlfile:
        if multi_coordinate:
            columns = pd.MultiIndex.from_tuples(coordinate,
                                                names=names or name)
        else:
            columns = pd.Index(coordinate, name=name or names)
        
        # Save data to excel
        data = pd.DataFrame(data=np.zeros([N_samples, N_coordinate]),
                            columns=columns)
        
        with pd.ExcelWriter(xlfile) as writer:
            for i, metric in zip(metrics, metric_data):
                data[:] = metric_data[metric]
                data.to_excel(writer, sheet_name=i.name)
    
    return metric_data
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 19:33:20 2018

@author: yoelr
"""
import numpy as np
from thermosteam import MultiStream, Stream
from .. import Unit
from ..utils import approx2step
from scipy.optimize import brentq
from ._hx import HXutility
import matplotlib.pyplot as plt
import biosteam as bst
array = np.array

# %% Equations

# Minimum thickness
x = array((4, 6, 8, 10, 12)) 
y = array((1/4, 5/16, 3/8, 7/16, 1/2))
ts_min_p = np.polyfit(x,y,1)

# %% Dictionary of factors

# Tray Type
F_TTdict = {'Sieve': 1,
            'Valve': 1.18,
            'Bubble cap': 1.87}

# Tray Materials (inner diameter, Di, in ft)

def compute_carbon_steel_material_factor(Di):
    return 1

def compute_stainless_steel_304_material_factor(Di):
    return 1.189 + 0.058*Di

def compute_stainless_steel_316_material_factor(Di):
    return 1.401 + 0.073*Di

def compute_carpenter_20CB3_material_factor(Di):
    return 1.525 + 0.079*Di

def compute_monel_material_factor(Di):
    return 2.306 + 0.112*Di

F_TMdict = {'Carbon steel': compute_carbon_steel_material_factor,
            'Stainless steel 304': compute_stainless_steel_304_material_factor,
            'Stainless steel 316': compute_stainless_steel_316_material_factor,
            'Carpenter 20CB-3': compute_carpenter_20CB3_material_factor,
            'Monel': compute_monel_material_factor}

# Column Material
F_Mdict = {'Carbon steel': 1.0,
           'Low-alloy steel': 1.2,
           'Stainless steel 304': 1.7,
           'Stainless steel 316': 2.1,
           'Carpenter 20CB-3': 3.2,
           'Nickel-200': 5.4,
           'Monel-400': 3.6,
           'Inconel-600': 3.9,
           'Incoloy-825': 3.7,
           'Titanium': 7.7}

# Material density (lb∕in^3)
rho_Mdict = {'Carbon steel': 0.284 ,
             'Low-alloy steel': None,
             'Stainless steel 304': 0.289,
             'Stainless steel 316': 0.289,
             'Carpenter 20CB-3': None,
             'Nickel-200': None,
             'Monel-400': None,
             'Inconel-600': None,
             'Incoloy-825': None,
             'Titanium': None}

# %% Distillation

class Dist(Unit, isabstract=True):
    """Abstract class for a column."""
    # Bare module factor
    BM = 4.3
    
    # Column material factor
    _F_Mstr = 'Carbon steel'
    _F_M = 1
    
    # Tray type factor
    _F_TTstr = 'Sieve'
    _F_TT = 1
    
    # Tray material factor function
    _F_TMstr = 'Carbon steel'
    _F_TM = staticmethod(F_TMdict['Carbon steel'])
    
    # [float] Tray spacing (225-600 mm)
    _TS = 450 
    
    #: [float] Enforced user defined stage efficiency.        
    _stage_efficiency = None
    
    # [float] Ratio of actual velocity to maximum velocity allowable before flooding.
    _f = 0.8 
    
    # [float] Foaming factor (0 < F_F < 1)
    _F_F = 1
    
    # [float] Ratio of open area, A_h, to active area, A_a
    _A_ha = 0.1
    
    # [float] Enforced ratio of downcomer area to net (total) area. If None, estimate ratio based on Oliver's estimation [1].
    _A_dn = None
    
    # [dict] Bounds for results
    _bounds = {'Diameter': (3., 24.),
               'Height': (27., 170.),
               'Weight': (9000., 2.5e6)}
    
    @property
    def LHK(self):
        return self._LHK
    @LHK.setter
    def LHK(self, LHK):
        # Set light non-key and heavy non-key indices
        self._LHK = LHK = tuple(LHK)
        chemicals = self.chemicals
        LHK_chemicals = LK_chemical, HK_chemical = self.chemicals.retrieve(LHK)
        Tb_light = LK_chemical.Tb
        Tb_heavy = HK_chemical.Tb
        LNK = []
        HNK = []
        if Tb_light > Tb_heavy:
            raise ValueError(f"light key must be lighter than heavy key")
        for chemical in chemicals:
            Tb = chemical.Tb
            if not Tb:
                HNK.append(chemical.ID)
            elif Tb < Tb_light:
                LNK.append(chemical.ID)
            elif Tb > Tb_heavy:
                HNK.append(chemical.ID)
            elif chemical not in LHK_chemicals:
                raise ValueError(f"intermediate volatile specie, '{chemical}', between light and heavy key, ['{LK_chemical}', '{HK_chemical}']")
        self._LNK = tuple(LNK)
        self._HNK = tuple(HNK)
    
    @property
    def y_top(self):
        return self._y_top
    @y_top.setter
    def y_top(self, y_top):
        self._y_top = y_top
        self._y = array([y_top, 1-y_top])
    
    @property
    def x_bot(self):
        return self._x_bot
    @x_bot.setter
    def x_bot(self, x_bot):
        self._x_bot = x_bot
        self._x = array([x_bot, 1-x_bot])
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 P=101325, *, LHK, y_top, x_bot, k):
        super().__init__(ID, ins, outs, thermo)
        self.P = P
        self.LHK = LHK
        self.y_top = y_top
        self.x_bot = x_bot
        self.k = k
    
    @property
    def tray_spacing(self):
        return self._TS
    @tray_spacing.setter
    def tray_spacing(self, TS):
        """Tray spacing (225-600 mm)."""
        self._TS = TS
    
    @property
    def stage_efficiency(self):
        """Enforced user defined stage efficiency."""
        return self._stage_efficiency
    @stage_efficiency.setter
    def stage_efficiency(self, stage_efficiency):
        self._stage_efficiency = stage_efficiency
    
    @property
    def velocity_factor(self):
        """Ratio of actual velocity to maximum velocity allowable before flooding."""
        return self._f
    @velocity_factor.setter
    def velocity_factor(self, f):
        self._f = f
    
    @property
    def foaming_factor(self):
        """Foaming factor (0 to 1)."""
        return self._F_F
    @foaming_factor.setter
    def foaming_factor(self, foaming_factor):
        if foaming_factor > 1 or foaming_factor < 0:
            raise ValueError(f"foaming_factor must be between 0 and 1, ({foaming_factor} given).")
        self._F_F = foaming_factor
    
    @property
    def open_area_factor(self):
        """Ratio of open area, A_h, to active area, A_a."""
        return self._A_ha
    @open_area_factor.setter
    def open_area_factor(self, A_ha):
        self._A_ha = A_ha
    
    @property
    def downcomer_area_factor(self):
        """Enforced ratio of downcomer area to net (total) area. If None, estimate ratio based on Oliver's estimation [1]."""
        return self._A_dn
    @downcomer_area_factor.setter
    def downcomer_area_factor(self, A_dn):
        self._A_dn = A_dn
    
    @property
    def tray_type(self):
        """Default 'Sieve'"""
        return self._F_TTstr
    @tray_type.setter
    def tray_type(self, tray_type):
        try:
            self._F_TT = F_TTdict[tray_type]
        except KeyError:
            dummy = str(F_TTdict.keys())[11:-2]
            raise ValueError(f"tray_type must be one of the following: {dummy}")
        self._F_TTstr = tray_type
        
    @property
    def tray_material(self):
        """Default 'Carbon steel'"""
        return self._F_TMstr
    @tray_material.setter
    def tray_material(self, tray_material):
        try:
            self._F_TM = F_TMdict[tray_material]
        except KeyError:
            dummy = str(F_TMdict.keys())[11:-2]
            raise ValueError(f"tray_material must be one of the following: {dummy}")
        self._F_TMstr = tray_material        

    @property
    def vessel_material(self):
        """Default 'Carbon steel'"""
        return self._F_Mstr
    @vessel_material.setter
    def vessel_material(self, vessel_material):
        try:
            self._F_M = F_Mdict[vessel_material]
        except KeyError:
            dummy = str(F_Mdict.keys())[11:-2]
            raise ValueError(f"vessel_material must be one of the following: {dummy}")
        self._F_Mstr = vessel_material  
    
    def _mass_balance(self):
        # Get all important flow rates (both light and heavy keys and non-keys)
        get_index = self.chemicals.get_index
        LHK_index = get_index(self._LHK)
        LNK_index = get_index(self._LNK)
        HNK_index = get_index(self._HNK)
        mol = self.mol_in
        LHK_mol = mol[LHK_index]
        HNK_mol = mol[HNK_index]
        LNK_mol = mol[LNK_index]

        # Set light and heavy keys by lever rule
        light, heavy = LHK_mol
        LHK_F_mol = light + heavy
        zf = light/LHK_F_mol
        split_frac = (zf-self.x_bot)/(self.y_top-self.x_bot)
        top_net = LHK_F_mol*split_frac

        # Set output streams
        vap, liq = self.outs
        vap.mol[LHK_index] = vap_LHK_mol = top_net * self._y
        liq.mol[LHK_index] = LHK_mol - vap_LHK_mol
        vap.mol[LNK_index] = LNK_mol
        liq.mol[HNK_index] = HNK_mol
    
    def _run(self):
        self._mass_balance()
        vap, liq = self.outs
        vap.phase = 'g'
        liq.phase = 'l'
        vap.P = liq.P = self.P
        self._condensate_dew_point = dp = vap.dew_point_at_P()
        self._boilup_bubble_point = bp = liq.bubble_point_at_P()
        liq.T = bp.T
        vap.T = dp.T

    def _equilibrium_staircase(self, operating_line, x_stairs,
                               y_stairs, T_stairs, x_limit, bubble_T):
        """Find the specifications at every stage of the of the operating line before the maximum liquid molar fraction. Append the light key liquid molar fraction, light key vapor molar fraction, and stage temperatures to x_stairs, y_stairs and T_stairs respectively.
        
        Parameters
        ----------
        operating_line : function
                         Should return the liquid molar fraction of the light
                         key given its vapor molar fraction.
        x_stairs : list
                   Liquid molar compositions at each stage. Last element
                   should be the starting point for the next stage.
        y_stairs : list
                   Vapor molar compositions at each stage. Last element 
                   should be the starting point for the next stage.
        T_stairs : list
                   Temperature at each stage.
            
        """
        P = self.P
        i = 0
        yi = y_stairs[-1]
        xi = x_stairs[-1]
        while xi < x_limit:
            if i > 100:
                raise RuntimeError('cannot meet specifications! stages > 100')
            i += 1
            # Go Up
            T, y = bubble_T(array((xi, 1-xi)), P)
            yi = y[0]
            y_stairs.append(yi)
            T_stairs.append(T)
            # Go Right
            xi = operating_line(yi)
            if xi > 1:
                xi = x_limit
            x_stairs.append(xi)

    def _plot_stages(self):
        """Plot stages, graphical aid line, and equilibrium curve. The plot does not include operating lines nor a legend."""
        vap, liq = self.outs
        if not hasattr(self, 'x_stages'):
            self._design()
        x_stages = self._x_stages
        y_stages = self._y_stages
        LHK = self.LHK
        LK = self.LHK[0]
        P = self.P
        
        # Equilibrium data
        x_eq = np.linspace(0, 1, 100)
        y_eq = np.zeros(100)
        T = np.zeros(100)
        n = 0
        
        bp = vap.get_bubble_point(IDs=LHK)
        solve_Ty = bp.solve_Ty
        for xi in x_eq:
            T[n], y = solve_Ty(array([xi, 1-xi]), P)
            y_eq[n] = y[0]
            n += 1
            
        # Set-up graph
        plt.figure()
        plt.xticks(np.arange(0, 1.1, 0.1), fontsize=12)
        plt.yticks(fontsize=12)
        plt.xlabel('x (' + LK + ')', fontsize=16)
        plt.ylabel('y (' + LK + ')', fontsize=16)
        plt.xlim([0, 1])
        
        # Plot stages
        x_stairs = []
        for x in x_stages:
            x_stairs.append(x)
            x_stairs.append(x)
            
        y_stairs = []
        for y in y_stages:
            y_stairs.append(y)
            y_stairs.append(y)
        x_stairs.pop(-1)
        x_stairs.insert(0, y_stairs[0])
        plt.plot(x_stairs, y_stairs, '--')
        
        # Graphical aid line
        plt.plot([0, 1], [0, 1])
        
        # Vapor equilibrium graph
        plt.plot(x_eq, y_eq, lw=2)

    def _cost_trays(self, N_T, Di):
        """Return total cost of all trays at a CE of 500 given number of trays, `N_T`, and inner diameter, `Di`, in ft."""
        # Note: Can only use this function after running design method.
        C_BT = self._calc_TrayBaseCost(Di)
        F_NT = self._calc_NTrayFactor(N_T)
        return N_T*F_NT*self._F_TT*self._F_TM(Di)*C_BT

    def _cost_tower(self, Di, L, W):
        """Return cost of tower at a CE of 500 given the inner diameter, `Di`, and length, `L` in ft and weight, `W`, in lb."""
        C_V = self._calc_EmptyTowerCost(W)
        C_PL = self._calc_PlaformLadderCost(Di, L)
        return (self._F_M*C_V + C_PL)
    
    @staticmethod
    def _calc_EmptyTowerCost(W):
        """Return C_V the cost (USD) of an empty tower vessel assuming a CE of 500.
        
        **Parameters**
        
            W: Weight (lb)
        
        """
        return np.exp(7.2756 + 0.18255*np.log(W) + 0.02297*np.log(W)**2)
    
    @staticmethod
    def _calc_PlaformLadderCost(Di, L):
        """Return C_PL, the cost (USD) of platforms and ladders assuming a CE of 500.
        
        **Parameters**
        
            Di: Inner diameter (ft)
            L: Legnth (ft)
        
        """
        return 300.9*Di**0.63316*L**0.80161
    
    @staticmethod
    def _calc_Weight(Di, L, tv, rho_M):
        """Return W, the weight (lb) of the tower assuming 2:1 elliptical head.
        
        **Parameters**
        
            Di: Diameter (ft)
            L: Legnth (ft)
            tv: Shell thickness (in)
            rho_M: Density of material (lb/in^3)
        
        """
        Di = Di*12
        L = L*12
        return np.pi*(Di+tv)*(L+0.8*Di)*tv*rho_M
    
    @staticmethod
    def _calc_WallThickness(Po, Di, L, S=15000, E=None, M=29.5):
        """Return tv, the wall thinkness (in) designed to withstand the internal pressure and the wind/earthquake load at the bottom.
        
        Parameters
        ----------
        Po : float
            Operating internal pressure (psi)
        Di : float
            Internal diameter (ft)
        L : float
            Height (ft)
        S : float
            Maximum stress (psi)
        E : float
            Fractional weld efficiency
        M : float
            Elasticity (psi)
            
        """
        # TODO: Incorporate temperature for choosing S and M
        Di = Di*12 # ft to in
        L = L*12
        
        E_check = E is None
        if E_check:
            # Assume carbon steel with thickness more than 1.25 in
            E = 1.0 
        
        # Get design pressure, which should be higher than operating pressure.
        Po_gauge = Po - 14.69
        if Po_gauge < 0:
            # TODO: Double check vacuum calculation
            Pd = -Po_gauge
            tE = 1.3*Di*(Pd*L/M/Di)**0.4
            tEC = L*(0.18*Di - 2.2)*10**-5 - 0.19
            tv = tE + tEC
            return tv
        elif Po_gauge < 5:
            Pd = 10
        elif Po_gauge < 1000:
            Pd = np.exp(0.60608 + 0.91615*np.log(Po)) + 0.0015655*np.log(Po)**2
        else:
            Pd = 1.1*Po_gauge
        
        # Calculate thinkess according to ASME pressure-vessel code.
        ts = Pd*Di/(2*S*E-1.2*Pd)
        
        if E_check:
            # Weld efficiency of 0.85 for low thickness carbon steel
            if ts < 1.25:
                E = 0.85
                ts = Pd*Di/(2*S*E-1.2*Pd)
        
        # Add corrosion allowence
        ts += 1/8
        
        # Minimum thickness for vessel rigidity may be larger
        Di_ft = Di/12
        ts_min = np.polyval(ts_min_p, Di/12) if Di_ft < 4 else 0.25
        if ts < ts_min:
            ts = ts_min
        
        # Calculate thickness to withstand wind/earthquake load
        Do = Di + ts
        tw = 0.22*(Do + 18)*L**2/(S*Do**2)
        tv = max(tw, ts)
        
        # Vessels are fabricated from metal plates with small increments
        if tv < 0.5:
            tv = approx2step(tv, 3/16, 1/16)
        elif tv < 2:
            tv = approx2step(tv, 0.5, 1/8)
        elif tv < 3:
            tv = approx2step(tv, 2, 1/4)
        return tv
    
    @staticmethod
    def _calc_TrayBaseCost(Di):
        """Return C_BT, the base cost of a tray (USD) at a CE of 500.
        
        **Parameters**
        
            Di: Inner diameter (ft)
        
        """
        return 412.6985 * np.exp(0.1482*Di)

    @staticmethod
    def _calc_NTrayFactor(N_T):
        """Return F_NT, the cost factor for number of trays.
        
        **Parameters**
        
            N_T: Number of trays
            
        """
        if N_T < 20:
            F_NT = 2.25/1.0414**N_T
        else:
            F_NT = 1
        return F_NT

    @staticmethod
    def _calc_MurphreeEfficiency(mu, alpha, L, V):
        """Return E_mv, the sectional murphree efficiency.
        
        **Parameters**
            
            mu: Viscosity (mPa*s)
            
            alpha: Relative volatility
            
            L: Liquid flow rate by mol
            
            V: Vapor flow rate by mol
        
        """
        S = alpha*V/L # Stripping factor
        e = 0.503*mu**(-0.226)*(S if S > 1 else 1/S)**(-0.08 )
        if e < 1: return e
        else: return 1
    
    @staticmethod
    def _calc_FlowParameter(L, V, rho_V, rho_L):
        """Return F_LV, the flow parameter.
        
        **Parameters**
        
            L: Liquid flow rate by mass
            V: Vapor flow rate by mass
            rho_V: Vapor density
            rho_L: Liquid density
        
        """
        return L/V*(rho_V/rho_L)**0.5
    
    @staticmethod
    def _calc_MaxCapacityParameter(TS, F_LV):
        """Return C_sbf, the maximum capacity parameter before flooding (m/s).
        
        **Parameters**
        
            TS :
                Tray spacing (mm)
            F_LV :
                Flow parameter
        
        """
        return 0.0105 + 8.127e-4*TS**0.755*np.exp(-1.463*F_LV**0.842)
    
    @staticmethod
    def _calc_MaxVaporVelocity(C_sbf, sigma,
                               rho_L, rho_V, F_F, A_ha):
        """Return U_f, the maximum allowable vapor velocity through the net area of flow before flooding (m/s).
        
        Parameters
        ----------
        
        C_sbf : 
            Maximum Capacity Parameter (m/s)
        sigma : 
            Liquid surface tension (dyn/cm)
        rho_L : 
            Liquid density
        rho_V : 
            Vapor density
        F_F : 
            Foaming factor
        A_ha : 
            Ratio of open area, A_h, to active area, A_a
        
        """
        F_ST = (sigma/20)**0.2 # Surface tension factor
        
        # Working area factor
        if A_ha >= 0.1 and A_ha <= 1:
            F_HA = 1
        elif A_ha >= 0.06:
            F_HA = 5*A_ha + 0.5
        else:
            raise ValueError(f"ratio of open to active area, 'A', must be between 0.06 and 1 ({A_ha} given)") 
        
        return C_sbf * F_HA * F_ST * ((rho_L-rho_V)/rho_V)**0.5
    
    @staticmethod
    def _calc_DowncomerAreaRatio(F_LV):
        """Return the ratio of downcomer area to net (total) area, `A_dn`.
        
        **Parameters**
        
            F_LV : float
                Flow parameter
        
        """
        if F_LV < 0.1:
            A_dn = 0.1
        elif F_LV < 1:
            A_dn = 0.1 + (F_LV-0.1)/9
        else:
            A_dn = 0.2
        return A_dn
    
    @staticmethod
    def _calc_Diameter(V_vol, U_f, f, A_dn):
        """Return D_T, the column diameter (ft).
        
        **Parameters**
        
            V_vol: Vapor volumetric flow rate (m^3/s)
            U_f: Maximum vapor velocity before flooding(m/s)
            f: Ratio of actual velocity to U_f
            A_dn: ratio of downcomer area to net (total) area
        
        """
        Di = (4*V_vol/(f*U_f*np.pi*(1-A_dn)))**0.5
        if Di < 0.914:
            # Make sure diameter is not too small
            Di = 0.914
        Di *= 3.28
        return Di
    
    @staticmethod
    def _calc_Height(TS, Nstages: int, top=True, bot=True):
        """Return H, the height of the column (ft).
        
        **Parameters**
        
            TS: Tray spacing (mm)
            Nstages: Number of stages 
        
        """
        # 3 m bottoms surge capacity, 1.25 m above top tray to remove entrained liquid
        H = TS*Nstages/1000
        if top:
            H += 1.2672
        if bot:
            H += 3
        H *= 3.28
        return H 

    def _calc_condenser(self):
        distillate = self.outs[0]
        condensate = self._condensate # Abstract attribute
        condenser = self._condenser
        s_in = condenser.ins[0]
        s_in.mol[:] = distillate.mol+condensate.mol
        s_in.T = distillate.T
        s_in.P = distillate.P
        ms1 = condenser.outs[0]
        ms1.imol['l'] = condensate.mol
        ms1.T = condensate.T
        ms1.P = condensate.P
        ms1.imol['g'] = distillate.mol
        condenser._design()
        condenser._cost()
        
    def _calc_boiler(self):
        bottoms = self.outs[1]
        boil_up = self._boil_up # Abstract attribute
        boiler = self._boiler
        s_in = boiler.ins[0]
        s_in.copy_like(bottoms)
        s_in.mol += boil_up.mol
        ms1 = boiler.outs[0]
        ms1.T = boil_up.T
        ms1.P = boil_up.P
        ms1.imol['g'] = boil_up.mol
        ms1.imol['l'] = bottoms.mol
        if hasattr(self, '_condenser'):
            boiler._design(self.H_out - self.H_in - self._condenser._duty)
            boiler._cost()
        else:
            boiler._design()
            boiler._cost()
        
    def _cost(self):
        Design = self.design_results
        Cost = self.purchase_costs
        F_CE = bst.CE/500
        
        # Cost trays assuming a partial condenser
        N_T = Design['Actual stages'] - 1
        Di = Design['Diameter']
        Cost['Trays'] = F_CE*self._cost_trays(N_T, Di)
        
        # Cost vessel assuming T < 800 F
        W = Design['Weight'] # in lb
        L = Design['Height']*3.28 # in ft
        Cost['Tower'] = F_CE*self._cost_tower(Di, L, W)
        self._cost_components()
        

class Distillation(Dist):
    """
    Create a Distillation column that assumes all light and heavy non keys
    separate to the top and bottoms product respectively. McCabe-Thiele
    analysis is used to find both the number of stages and the reflux ratio
    given a ratio of actual reflux to minimum reflux [1]_. This assumption
    is good for both binary distillation of highly polar compounds and
    ternary distillation assuming complete separation of light non-keys
    and heavy non-keys with large differences in boiling points. Preliminary
    analysis showed that the theoretical number of stages using this method
    on Methanol/Glycerol/Water systems is off by less than +-1 stage. Other
    methods, such as the Fenske-Underwood-Gilliland method, are more suitable
    for hydrocarbons. The Murphree efficiency is based on the modified
    O'Connell correlation [2]_. The diameter is based on tray separation
    and flooding velocity [1]_. Purchase costs are based on correlations
    by Mulet et al. [3, 4]_ as compiled by Warren et. al. [5]_.

    Parameters
    ----------
    ins
        [:] All input streams
    outs
        [0] Distillate product
        
        [1] Bottoms product
    LHK : tuple[str]
          Light and heavy keys.
    P=101325 : float
        Operating pressure (Pa).
    y_top : float
            Molar fraction of light key in the distillate.
    x_bot : float
            Molar fraction of light key in the bottoms.
    k : float
        Ratio of reflux to minimum reflux.

    References
    ----------
    .. [1] J.D. Seader, E.J. Henley, D.K. Roper. (2011)
        Separation Process Principles 3rd Edition. John Wiley & Sons, Inc. 

    .. [2] M. Duss, R. Taylor. (2018)
        Predict Distillation Tray Efficiency. AICHE 
    
    .. [3] Mulet, A., A. B. Corripio, and L. B. Evans. (1981a).
        Estimate Costs of Pressure Vessels via Correlations.
        Chem. Eng., 88(20), 145–150.

    .. [4] Mulet, A., A.B. Corripio, and L.B.Evans. (1981b).
        Estimate Costs of Distillation and Absorption Towers via Correlations.
        Chem. Eng., 88(26), 77–82.

    .. [5] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R.,
        & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
        Cost Accounting and Capital Cost Estimation (Chapter 16)    

    Examples
    --------
    Binary distillation assuming 100% separation on non-keys:
    
    >>> from biosteam.units import Distillation
    >>> from thermosteam import Stream, Chemicals, settings
    >>> chemicals = Chemicals(['Water', 'Methanol', 'Glycerol'])
    >>> settings.set_thermo(chemicals)
    >>> feed = Stream('feed', flow=(80, 100, 25))
    >>> bp = feed.bubble_point_at_P()
    >>> feed.T = bp.T # Feed at bubble point T
    >>> D1 = Distillation('D1', ins=feed,
    ...                   LHK=('Methanol', 'Water'),
    ...                   y_top=0.99, x_bot=0.01, k=2)
    >>> D1.is_divided = True
    >>> D1.simulate()
    >>> # See all results
    >>> D1.show(T='degC', P='atm', composition=True)
    Distillation: D1
    ins...
    [0] feed
        phase: 'l', T: 76.129 degC, P: 1 atm
        composition: Water     0.39
                     Methanol  0.488
                     Glycerol  0.122
                     --------  205 kmol/hr
    outs...
    [0] s1
        phase: 'g', T: 64.91 degC, P: 1 atm
        composition: Water     0.01
                     Methanol  0.99
                     --------  100 kmol/hr
    [1] s2
        phase: 'l', T: 100.06 degC, P: 1 atm
        composition: Water     0.754
                     Methanol  0.00761
                     Glycerol  0.239
                     --------  105 kmol/hr
    
    >>> D1.results()
    Distillation                                    Units        D1
    Power               Rate                           kW         0
                        Cost                       USD/hr         0
    Cooling water       Duty                        kJ/hr -5.11e+06
                        Flow                      kmol/hr  3.49e+03
                        Cost                       USD/hr      1.71
    Low pressure steam  Duty                        kJ/hr  9.49e+06
                        Flow                      kmol/hr       244
                        Cost                       USD/hr      58.1
    Design              Theoretical feed stage                    9
                        Theoretical stages                       13
                        Minimum reflux              Ratio     0.687
                        Reflux                      Ratio      1.37
                        Rectifier stages                         14
                        Stripper stages                          13
                        Rectifier height               ft      33.2
                        Stripper height                ft      31.7
                        Rectifier diameter             ft      3.93
                        Stripper diameter              ft      3.01
                        Rectifier wall thickness       in      0.25
                        Stripper wall thickness        in      0.25
                        Rectifier weight               lb  4.61e+03
                        Stripper weight                lb  3.32e+03
    Purchase cost       Rectifier trays               USD  1.45e+04
                        Stripper trays                USD  1.21e+04
                        Rectifier tower               USD  7.41e+04
                        Stripper tower                USD   6.1e+04
                        Condenser                     USD  3.41e+04
                        Boiler                        USD  2.64e+04
    Total purchase cost                               USD  2.22e+05
    Utility cost                                   USD/hr      59.8
    
    """
    line = 'Distillation'
    _N_heat_utilities = 0
    _graphics = Dist._graphics
    _is_divided = False #: [bool] True if the stripper and rectifier are two separate columns.
    _units_not_divided = {'Minimum reflux': 'Ratio',
                          'Reflux': 'Ratio',
                          'Rectifier height': 'ft',
                          'Rectifier diameter': 'ft',
                          'Rectifier wall thickness': 'in',
                          'Rectifier weight': 'lb',
                          'Stripper height': 'ft',
                          'Stripper diameter': 'ft',
                          'Stripper wall thickness': 'in',
                          'Stripper weight': 'lb',
                          'Height': 'ft',
                          'Diameter': 'ft',
                          'Wall thickness': 'in',
                          'Weight': 'lb'}
    _units = _units_not_divided
    _units_divided = {'Minimum reflux': 'Ratio',
                      'Reflux': 'Ratio',
                      'Rectifier height': 'ft',
                      'Rectifier diameter': 'ft',
                      'Rectifier wall thickness': 'in',
                      'Rectifier weight': 'lb',
                      'Stripper height': 'ft',
                      'Stripper diameter': 'ft',
                      'Stripper wall thickness': 'in',
                      'Stripper weight': 'lb'}
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 P=101325, *, LHK, y_top, x_bot, k):
        super().__init__(ID, ins, outs, thermo, P, LHK=LHK,
                         y_top=y_top, x_bot=x_bot, k=k)
        thermo = self._load_thermo(thermo)
        self._condenser = HXutility(None,
                                    ins=Stream(None, phase='g', thermo=thermo),
                                    outs=MultiStream(None, thermo=thermo),
                                    thermo=thermo)
        self._boiler = HXutility(None,
                                 ins=Stream(None, thermo=thermo),
                                 outs=MultiStream(None, thermo=thermo),
                                 thermo=thermo)
        self.heat_utilities = self._condenser.heat_utilities + self._boiler.heat_utilities
        self._condensate = Stream(None, thermo=thermo)
        self._boil_up = Stream(None, phase='g', thermo=thermo)
        self._vapor_stream =  Stream(None)
        self._McCabeThiele_args = np.zeros(6)
    
    @property
    def is_divided(self):
        """[bool] True if the stripper and rectifier are two separate columns."""
        return self._is_divided
    
    @is_divided.setter
    def is_divided(self, is_divided):
        self._is_divided = is_divided
        self._units = self._units_divided if is_divided else self._units_not_divided
    
    def _calc_Nstages(self):
        """Return a tuple with the actual number of stages for the rectifier and the stripper."""
        vap, liq = self.outs
        Design = self.design_results
        x_stages = self._x_stages
        y_stages = self._y_stages
        R = Design['Reflux']
        N_stages = Design['Theoretical stages']
        feed_stage = Design['Theoretical feed stage']
        liq_mol = self._feed_liqmol
        vap_mol = self._feed_vapmol
        
        stage_efficiency = self.stage_efficiency
        if stage_efficiency:
            return N_stages/stage_efficiency
        else:    
            # Calculate Murphree Efficiency for rectifying section
            vap_F_mol = vap.F_mol
            dp = self._condensate_dew_point
            condensate_x_mol = dp.x
            self._L_Rmol = L_Rmol = R*vap_F_mol
            self._V_Rmol = V_Rmol = (R+1)*vap_F_mol
            condensate = self._condensate
            condensate.imol[dp.IDs] = condensate_x_mol * L_Rmol
            condensate.T = vap.T
            condensate.P = vap.P
            mu = 1000 * condensate.mu # mPa*s
            K_light = y_stages[-1]/x_stages[-1] 
            K_heavy = (1-y_stages[-1])/(1-x_stages[-1])
            alpha = K_light/K_heavy
            self._E_rectifier = E_rectifier = self._calc_MurphreeEfficiency(mu, alpha, L_Rmol, V_Rmol)
            
            # Calculate Murphree Efficiency for stripping section
            mu = 1000 * liq.mu # mPa*s
            self._V_Smol = V_Smol = (R+1)*vap_F_mol - sum(vap_mol)
            self._L_Smol = L_Smol = R*vap_F_mol + sum(liq_mol) 
            K_light = y_stages[0]/x_stages[0] 
            K_heavy = (1-y_stages[0])/(1-x_stages[0] )
            alpha = K_light/K_heavy
            self._E_stripper = E_stripper = self._calc_MurphreeEfficiency(mu, alpha, L_Smol, V_Smol)
            
            # Calculate actual number of stages
            mid_stage = feed_stage - 0.5
            N_rectifier = np.ceil(mid_stage/E_rectifier)
            N_stripper = np.ceil( (N_stages-mid_stage)/E_stripper )
        return N_rectifier, N_stripper

    def _McCabeThiele(self):
        distillate, bottoms = self.outs
        chemicals = self.chemicals
        LHK = self._LHK
        LHK_index = chemicals.get_index(LHK)

        # Feed light key mol fraction
        liq_mol = np.zeros(chemicals.size)
        vap_mol = liq_mol.copy()
        for s in self.ins:
            if isinstance(s, MultiStream):
                liq_mol += s.imol['l']
                vap_mol += s.imol['g']
            elif s.phase == 'g':
                vap_mol += s.mol
            elif s.phase.lower() == 'l':
                liq_mol += s.mol
            elif s.phase == 's': pass
            else:
                raise RuntimeError(f'invalid phase encountered in {repr(s)}')
        self._feed_liqmol = liq_mol
        self._feed_vapmol = vap_mol
        LHK_mol = liq_mol[LHK_index] + vap_mol[LHK_index]
        LHK_F_mol = LHK_mol.sum()
        zf = LHK_mol[0]/LHK_F_mol
        
        # Get feed quality
        q = liq_mol[LHK_index].sum()/LHK_F_mol
        
        # Main arguments
        P = self.P
        k = self.k
        y_top = self.y_top
        x_bot = self.x_bot
        
        # Cache
        old_args = self._McCabeThiele_args
        args = np.array([P, k, y_top, x_bot, q, zf])
        if (abs(old_args - args) < np.array([50, 1e-5, 1e-6, 1e-6, 1e-6, 1e-6], float)).all(): return
        self._McCabeThiele_args = args
        
        # Get R_min and the q_line 
        if q == 1:
            q = 1 - 1e-5
        q_line = lambda x: q*x/(q-1) - zf/(q-1)
        self._q_line_args = dict(q=q, zf=zf)
        
        solve_Ty = bottoms.get_bubble_point(LHK).solve_Ty
        Rmin_intersection = lambda x: q_line(x) - solve_Ty(array((x, 1-x)), P)[1][0]
        x_Rmin = brentq(Rmin_intersection, 0, 1)
        y_Rmin = q_line(x_Rmin)
        m = (y_Rmin-y_top)/(x_Rmin-y_top)
        Rmin = m/(1-m)
        if Rmin <= 0:
            R = 0.0001*k
        else:
            R = Rmin*k

        # Rectifying section: Inntersects q_line with slope given by R/(R+1)
        m1 = R/(R+1)
        b1 = y_top-m1*y_top
        rs = lambda y: (y - b1)/m1 # -> x
        
        # y_m is the solution to lambda y: y - q_line(rs(y))
        self._y_m = y_m = (q*b1 + m1*zf)/(q - m1*(q-1))
        self._x_m = x_m = rs(y_m)
        
        # Stripping section: Intersects Rectifying section and q_line and beggins at bottoms liquid composition
        m2 = (x_bot-y_m)/(x_bot-x_m)
        b2 = y_m-m2*x_m
        ss = lambda y: (y-b2)/m2 # -> x        
        
        # Data for staircase
        self._x_stages = x_stages = [x_bot]
        self._y_stages = y_stages = [x_bot]
        self._T_stages = T_stages = []
        self._equilibrium_staircase(ss, x_stages, y_stages, T_stages, x_m, solve_Ty)
        yi = y_stages[-1]
        xi = rs(yi)
        x_stages[-1] = xi if xi < 1 else 0.99999
        self._equilibrium_staircase(rs, x_stages, y_stages, T_stages, y_top, solve_Ty)
        
        # Find feed stage
        for i in range(len(y_stages)-1):
            j = i+1
            if y_stages[i] < y_m and y_stages[j] > y_m:
                feed_stage = i+1
        stages = len(x_stages)
        
        # Results
        Design = self.design_results
        Design['Theoretical feed stage'] = stages - feed_stage
        Design['Theoretical stages'] = stages
        Design['Minimum reflux'] = Rmin
        Design['Reflux'] = R
        

    def _design(self):
        self._McCabeThiele()
        distillate, bottoms = self._outs
        Design = self.design_results
        R = Design['Reflux']
        Rstages, Sstages = self._calc_Nstages()
        calc_Height = self._calc_Height
        is_divided = self.is_divided
        TS = self._TS
        
        ### Get diameter of rectifying section based on top plate ###
        
        condensate = self._condensate
        rho_L = condensate.rho
        sigma = 1000 * condensate.sigma # dyn/cm
        L = condensate.F_mass
        V = L*(R+1)/R
        vapor_stream = self._vapor_stream
        vapor_stream.copy_like(distillate)
        vapor_stream.mol *= R+1
        V_vol = 0.0002778 * vapor_stream.F_vol # m^3/s
        rho_V = distillate.rho
        F_LV = self._calc_FlowParameter(L, V, rho_V, rho_L)
        C_sbf = self._calc_MaxCapacityParameter(TS, F_LV)
        F_F = self._F_F
        A_ha = self._A_ha
        U_f = self._calc_MaxVaporVelocity(C_sbf, sigma, rho_L, rho_V, F_F, A_ha)
        A_dn = self._A_dn
        if A_dn is None:
           self._A_dn = A_dn = self._calc_DowncomerAreaRatio(F_LV)
        f = self._f
        R_diameter = self._calc_Diameter(V_vol, U_f, f, A_dn)
        
        ### Get diameter of stripping section based on feed plate ###
        
        V_mol = self._V_Smol
        rho_L = bottoms.rho
        bp = self._boilup_bubble_point
        boil_up_flow = bp.y * V_mol
        boil_up = self._boil_up
        boil_up.T = bottoms.T
        boil_up.P = bottoms.P
        boil_up.imol[bp.IDs] = boil_up_flow
        V = boil_up.F_mass
        V_vol = 0.0002778 * boil_up.F_vol # m^3/s
        rho_V = boil_up.rho
        L = bottoms.F_mass # To get liquid going down
        F_LV = self._calc_FlowParameter(L, V, rho_V, rho_L)
        C_sbf = self._calc_MaxCapacityParameter(TS, F_LV)
        sigma = 1000 * bottoms.sigma
        U_f = self._calc_MaxVaporVelocity(C_sbf, sigma, rho_L, rho_V, F_F, A_ha)
        A_dn = self._A_dn
        if A_dn is None:
            self.A_dn = self._calc_DowncomerAreaRatio(F_LV)
        S_diameter = self._calc_Diameter(V_vol, U_f, f, A_dn)
        Po = self.P*0.000145078 # to psi
        rho_M = rho_Mdict[self._F_Mstr]
        
        if is_divided:
            Design['Rectifier stages'] = Rstages
            Design['Stripper stages'] =  Sstages
            Design['Rectifier height'] = H_R = calc_Height(TS, Rstages-1)
            Design['Stripper height'] = H_S = calc_Height(TS, Sstages-1)
            Design['Rectifier diameter'] = R_diameter
            Design['Stripper diameter'] = S_diameter
            Design['Rectifier wall thickness'] = tv = self._calc_WallThickness(Po, R_diameter, H_R)
            Design['Stripper wall thickness'] = tv = self._calc_WallThickness(Po, S_diameter, H_S)
            Design['Rectifier weight'] = self._calc_Weight(R_diameter, H_R, tv, rho_M)
            Design['Stripper weight'] = self._calc_Weight(S_diameter, H_S, tv, rho_M)
        else:
            Design['Actual stages'] = Rstages + Sstages
            Design['Height'] = H = calc_Height(TS, Rstages+Sstages-2)
            Design['Diameter'] = Di = max((R_diameter, S_diameter))
            Design['Wall thickness'] = tv = self._calc_WallThickness(Po, Di, H)
            Design['Weight'] = self._calc_Weight(Di, H, tv, rho_M)
        
    def _cost(self):
        if not self.is_divided: return super()._cost()
        Design = self.design_results
        Cost = self.purchase_costs
        F_CE = bst.CE/500
        
        # Number of trays assuming a partial condenser
        N_RT = Design['Rectifier stages'] - 1
        Di_R = Design['Rectifier diameter']
        Cost['Rectifier trays'] = F_CE*self._cost_trays(N_RT, Di_R)
        N_ST = Design['Stripper stages'] - 1
        Di_S = Design['Stripper diameter']
        Cost['Stripper trays'] = F_CE*self._cost_trays(N_ST, Di_S)
        
        # Cost vessel assuming T < 800 F
        W_R = Design['Rectifier weight'] # in lb
        H_R = Design['Rectifier height']*3.28 # in ft
        Cost['Rectifier tower'] = F_CE*self._cost_tower(Di_R, H_R, W_R)
        W_S = Design['Stripper weight'] # in lb
        H_S = Design['Stripper height']*3.28 # in ft
        Cost['Stripper tower'] = F_CE*self._cost_tower(Di_S, H_S, W_S)
        self._cost_components()
    
    def _cost_components(self): 
        # Cost condenser
        self._calc_condenser()
        self.purchase_costs['Condenser'] = self._condenser.purchase_costs['Heat exchanger']
        
        # Cost boiler
        self._calc_boiler()
        self.purchase_costs['Boiler'] = self._boiler.purchase_costs['Heat exchanger']
        
    
    def plot_stages(self):
        """Plot the McCabe Thiele Diagram."""
        # Plot stages, graphical aid and equilibrium curve
        self._plot_stages()
        vap, liq = self.outs
        Design = self.design_results
        if not hasattr(self, '_x_stages'): self._design()
        q_args = self._q_line_args
        zf = q_args['zf']
        q = q_args['q']
        q_line = lambda x: q*x/(q-1) - zf/(q-1)
        y_top = self.y_top
        x_bot = self.x_bot
        stages = Design['Theoretical stages']
        Rmin = Design['Minimum reflux']
        R = Design['Reflux']
        feed_stage = Design['Theoretical feed stage']
        
        # q_line
        intersect2 = lambda x: x - q_line(x)
        x_m2 = brentq(intersect2, 0, 1)
        
        # Graph q-line, Rectifying and Stripping section
        plt.plot([self._x_m, x_m2], [self._y_m, x_m2])
        plt.plot([self._x_m, y_top], [self._y_m, y_top])
        plt.plot([x_bot, self._x_m], [x_bot, self._y_m])
        plt.legend([f'Stages: {stages}, Feed: {feed_stage}', 'Graphical aid', 'eq-line', 'q-line', 'ROL', 'SOL'], fontsize=12)
        plt.title(f'McCabe Thiele Diagram (Rmin = {Rmin:.2f}, R = {R:.2f})')
        plt.show()
        return plt


# class Stripper(Dist):
#     line = 'Stripper'
#     __doc__ = column_doc.replace('{Column Type}', 'Stripper')
#     _N_heat_utilities = 0
#     _graphics = Dist._graphics
#     _units ={'Minimum boil up': 'Ratio',
#              'Boil up': 'Ratio',
#              'Height': 'ft',
#              'Diameter': 'ft',
#              'Wall thickness': 'in',
#              'Weight': 'lb'}
    
#     def _init(self):
#         self._boiler = HXutility(None,
#                                  ins=Stream(None),
#                                  outs=MixedStream(None))
#         self.heat_utilities = self._boiler.heat_utilities
#         self._boil_up = Stream(None)
#         self._McCabeThiele_args = np.array([0, 0, 0, 0])
    
#     def plot_stages(self):
#         # Plot stages, graphical aid and equilibrium curve
#         self._plot_stages()
        
#         # Cached Data
#         vap, liq = self.outs
#         Design = self.design_results
#         cached = self._cached
#         x_stages = cached.get('x_stages')
#         if not x_stages:
#             self._design()
#             x_stages = cached.get('x_stages')
#         y_top = self.y_top
#         x_bot = self.x_bot
#         stages = Design['Theoretical stages']
#         Bmin = Design['Minimum boil up']
#         B = Design['Boil up']
        
#         # Graph Stripping section
#         ss = self._ss
#         plt.plot([x_bot, ss(y_top)], [x_bot, y_top])
#         plt.legend([f'Stages: {stages}', 'Graphical aid', 'eq-line', 'SOL'], fontsize=12)
            
#         # Title
#         plt.title(f'McCabe Thiele Diagram (Bmin = {Bmin:.2f}, B = {B:.2f})')
#         plt.show()
#         return plt
    
#     def _calc_Nstages(self):
#         """Return the actunal number of stages"""
#         vap, liq = self.outs
#         cached = self._cached
#         Design = self.design_results
#         x_stages = cached['x_stages']
#         y_stages = cached['y_stages']
#         LHK_index = cached['LHK_index']
#         B = Design['Boil up']
#         N_stages = Design['Theoretical stages']
        
#         stage_efficiency =self.stage_efficiency
#         if stage_efficiency:
#             return N_stages/stage_efficiency
#         else:
#             # Calculate Murphree Efficiency for stripping section
#             mu = 1000 * liq.mu # mPa*s
#             cached['V_mol'] = V_mol = B*sum(liq.mol[LHK_index])
#             cached['L_mol'] = L_mol = liq.F_mol + V_mol
#             K_light = y_stages[0]/x_stages[0] 
#             K_heavy = (1-y_stages[0])/(1-x_stages[0] )
#             alpha = K_light/K_heavy
#             cached['Stripping Section Efficiency'] = E_stripper = self._calc_MurphreeEfficiency(mu, alpha, L_mol, V_mol)
#         return np.ceil(N_stages/E_stripper)
    
#     def _McCabeThiele(self):
#         distillate, bottoms = self.outs
#         Design = self.design_results
#         cached = self._cached
        
#         # Some important info
#         species = self._LHK_chemicals
#         P = self.P
#         k = self.k
#         y_top = self.y_top
#         x_bot = self.x_bot
        
#         old_args = self._McCabeThiele_args
#         args = np.array([y_top, x_bot, k, P])
#         if (abs(old_args - args) < np.array([0.00001, 0.00001, 0.01, 50], float)).all(): return
#         self._McCabeThiele_args = args
        
#         # Get B_min (Boil up ratio)
#         y_Rmin = y_top
#         T_guess, x = distillate._dew_T(species, [y_Rmin, 1-y_Rmin], P)
#         x_Rmin = x[0]
#         m = (y_Rmin-x_bot)/(x_Rmin-x_bot)
#         Bmin = 1/(m-1)
#         B = 0.1*k if Bmin <= 0 else Bmin*k

#         # Stripping section: Inntersects liquid composition with slope given by (B+1)/B
#         m = (B+1)/B
#         b = x_bot-m*x_bot
#         def ss(y): return (y - b)/m
#         self._ss = ss
                
#         # Data for staircase
#         cached['x_stages'] = x_stages = [x_bot]
#         cached['y_stages'] = y_stages = [x_bot]
#         cached['T_stages'] = T_stages = []
#         self._equilibrium_staircase(ss, x_stages, y_stages, T_stages, ss(y_top))
#         stages = len(x_stages)
        
#         # Results
#         Design['Theoretical stages'] = stages
#         Design['Minimum boil up'] = Bmin
#         Design['Boil up'] = B
    
#     def _design(self):
#         distillate, bottoms = self.outs
#         Design = self.design_results
#         cached = self._cached
        
#         ### Get number of stages and height ###
        
#         self._McCabeThiele()
#         TS = self._TS
#         Design['Actual stages'] = Nstages = self._calc_Nstages()
#         Design['Height'] = H = self._calc_Height(TS, Nstages-1)
        
#         ### Get diameter of stripping section based on feed plate ###
        
#         V_mol = cached['V_mol']
#         L_mol = cached['L_mol']
#         rho_L = bottoms.rho
#         boil_up_flow = cached['boilup_z_mol'] * V_mol
#         boil_up = cached['boil_up']
#         boil_up.T = bottoms.T; boil_up.P = bottoms.P
#         lookup = bottoms._species._compounds.index
#         index_ = [lookup(i) for i in cached['vle_bot']]
#         boil_up.mol[index_] = boil_up_flow
#         V = boil_up.F_mass
#         V_vol = 0.0002778 * boil_up.F_vol # m^3/s
#         rho_V = boil_up.rho
#         L = sum(bottoms.MW*bottoms.z_mol*L_mol) # To get liquid going down
#         F_LV = self._calc_FlowParameter(L, V, rho_V, rho_L)
#         C_sbf = self._calc_MaxCapacityParameter(TS, F_LV)
#         sigma = 1000 * bottoms.sigma # dyn/cm
#         U_f = self._calc_MaxVaporVelocity(C_sbf, sigma, rho_L, rho_V, self._F_F, self._A_ha)
#         A_dn = self._A_dn or self._calc_DowncomerAreaRatio(F_LV)
#         Design['Diameter'] = Di = self._calc_Diameter(V_vol, U_f, self._f, A_dn)
#         Po = self.P/101325*14.7
#         Design['Wall thickness'] = tv = self._calc_WallThickness(Po, Di, H)
#         rho_M = rho_Mdict[self._F_Mstr]
#         Design['Weight'] = self._calc_Weight(Di, H, tv, rho_M)
    
#     def _cost_components(self):
#         # Cost boiler
#         self._calc_boiler()
#         self.purchase_costs['Boiler'] = self._boiler.purchase_costs['Heat exchanger']

    


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 18 07:19:49 2019

@author: yoelr
"""
from warnings import warn

__all__ = ('DesignWarning', 'design_warning',
           'lb_warning', 'ub_warning', 'bounds_warning')

class DesignWarning(Warning):
    """Warning regarding design constraints."""

# %% Bounds checking

def design_warning(source, msg):
    """Return a DesignWarning object with source description."""
    if isinstance(source, str):
        msg = f'@{source}: ' + msg
    elif source:
        msg = f'@{type(source).__name__} {str(source)}: ' + msg
    return DesignWarning(msg)
            
def lb_warning(key, value, units, lb, stacklevel, source):
    units = ' ' + units if units else ''
    try:
        msg = f"{key} ({value:.4g}{units}) is out of bounds (minimum {lb:.4g}{units})."
    except:  # Handle format errors
        msg = f"{key} ({value:.4g}{units}) is out of bounds (minimum {lb}{units})."
    
    warn(design_warning(source, msg, DesignWarning), stacklevel=stacklevel)
    
def ub_warning(key, value, units, ub, stacklevel, source):
    units = ' ' + units if units else ''
    try:
        msg = f"{key} ({value:.4g}{units}) is out of bounds (maximum {ub:.4g}{units})."
    except:  # Handle format errors
        msg = f"{key} ({value:.4g}{units}) is out of bounds (maximum {ub}{units})."
    
    warn(design_warning(source, msg, DesignWarning), stacklevel=stacklevel)

def bounds_warning(source, key, value, units, bounds):
    """Issue a warning if value is out of bounds.
    
    Parameters
    ----------
    key : str
          Name of value.
    value : float
    units : str
            Units of value        
    bounds : iterable[float, float]
             Upper and lower bounds.
        
    """
    # Warn when value is out of bounds
    lb, ub = bounds
    if not (lb <= value <= ub):
        units = ' ' + units if units else ''
        try:
            msg = f"{key} ({value:.4g}{units}) is out of bounds ({lb:.4g} to {ub:.4g}{units})."
        except:  # Handle format errors
            msg = f"{key} ({value:.4g}{units}) is out of bounds ({lb} to {ub}{units})."
        warn(design_warning(source, msg), stacklevel=3)


# -*- coding: utf-8 -*-
"""
Created on Sun Feb 17 16:36:47 2019

@author: yoelr
"""
from thermosteam import Stream
from ... import HeatUtility, Unit
from . import Facility
from ..decorators import cost

__all__ = ('BoilerTurbogenerator',)

lps = HeatUtility.heating_agents['Low pressure steam']

#: TODO add reference of NREL

@cost('Work', 'Turbogenerator',
      CE=551, S=42200, kW=0, cost=9500e3, n=0.60, BM=1.8)
@cost('Flow rate', 'Hot process water softener system', 
      CE=551, cost=78e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Amine addition pkg', 
      CE=551, cost=40e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Deaerator',
      CE=551, cost=305e3, S=235803, n=0.6, BM=3.0)
@cost('Flow rate', 'Boiler',
      CE=551, cost=28550e3, kW=1000, S=238686, n=0.6, BM=1.8)
class BoilerTurbogenerator(Facility):
    """Create a BoilerTurbogenerator object that will calculate electricity generation from burning the feed. It also takes into account how much steam is being produced, and the required cooling utility of the turbo generator. No emissions or mass balances are taken into account.
    
    Parameters
    ----------
    ins
        [0] Feed that will be burned.
    outs
        [0] Emission (burned feed)
    boiler_efficiency : float
        Fraction of heat transfered to steam.
    turbo_generator_efficiency : float
        Fraction of steam heat converted to electricity.
    """
    #: TODO: Make this a whole system instead of approximating duty per mol values
    duty_over_mol = 45000 # Superheat steam with 45000 kJ/kmol
    boiler_blowdown = 0.03
    RO_rejection = 0
    _N_outs = _N_ins = 3
    _N_heat_utilities = 2
    _has_power_utility = True
    _units = {'Flow rate': 'kg/hr',
              'Work': 'kW'}
    
    def __init__(self, ID='', ins=None, outs=(), *,
                 boiler_efficiency=0.80,
                 turbogenerator_efficiency=0.85,
                 side_steam=None):
        Unit.__init__(self, ID, ins, outs)
        self.makeup_water = makeup_water = Stream('boiler_makeup_water', thermo=lps.thermo)
        loss = makeup_water.flow_proxy()
        loss.ID = 'rejected_water_and_blowdown'
        self.ins[-1] = makeup_water
        self.outs[1] = loss
        self.boiler_efficiency = boiler_efficiency
        self.turbogenerator_efficiency = turbogenerator_efficiency
        self.steam_utilities = set()
        self.steam_demand = Stream(None, thermo=lps.thermo, 
                                   P=lps.P, T=lps.T, phase='g')
        self.side_steam = side_steam
    
    def _run(self): pass
    
    def _design(self):
        B_eff = self.boiler_efficiency
        TG_eff = self.turbogenerator_efficiency
        steam = self.steam_demand
        steam_utilities = self.steam_utilities
        if not steam_utilities:
            for u in self.system.units:
                if u is self: continue
                for hu in u.heat_utilities:
                    if hu.ID == 'Low pressure steam':
                        steam_utilities.add(hu)
        steam.imol['7732-18-5'] = steam_mol = sum([i.flow for i in steam_utilities])
        duty_over_mol = self.duty_over_mol
        feed = self._ins[0]
        feed2 = self._ins[1]
        emission = self._outs[0]
        hu_cooling, hu_steam = self.heat_utilities
        H_steam = steam_mol * duty_over_mol
        if self.side_steam: 
            H_steam += self.side_steam.H
        feed_Hc = feed.Hc
        if feed2:
            feed_Hc += feed2.Hc
        
        # This is simply for the mass balance (no special purpose)
        emission.mol[:] = feed.mol # TODO: In reality, this should be CO2
        
        # A percent of bagasse is water, so remove latent heat of vaporization
        F_mass_feed = feed.F_mass
        moisture_content = feed.imass['7732-18-5']/F_mass_feed
        H_content = feed_Hc*B_eff - F_mass_feed*2300*moisture_content
        
        #: [float] Total steam produced by the boiler (kmol/hr)
        self.total_steam = H_content / 40000 
        # Note: A portion of the steam produced is at milder conditions,
        #       so it does not consume as much energy.
        #       This is a really vague approximation, a more rigorous 
        #       model is needed (i.e. simulate whole system).
        
        self.makeup_water.mol[0] = makeup_mol = self.total_steam * self.boiler_blowdown * 1/(1-self.RO_rejection)
        
        # Heat available for the turbogenerator
        H_electricity = H_content - H_steam - makeup_mol * 18000
        
        Design = self.design_results
        Design['Flow rate'] = self.total_steam * 18.01528
        
        if H_electricity < 0:
            H_steam = H_content
            cooling = electricity = 0
        else:
            electricity = H_electricity * TG_eff
            cooling = electricity - H_electricity
        hu_cooling(cooling, steam.T)
        hu_steam.ID = 'Low pressure steam'
        hu_steam.cost = -sum([i.cost for i in steam_utilities])
        Design['Work'] = electricity/3600

    def _end_decorated_cost_(self):
        self.power_utility(self.power_utility.rate - self.design_results['Work'])

# Simulation of ethanol production from sugarcane
# in Brazil: economic study of an autonomous
# distillery
# Marina O.S. Diasa,b, Marcelo P. Cunhaa, Charles D.F. Jesusa, Mirna I.G.
# Scandiffioa, Carlos E.V. Rossella,b, Rubens Maciel Filhob, Antonio Bonomia
# a CTBE – Bioethanol Science and Technology National Laboratory, PO Box 6170 –
# CEP 13083-970, Campinas – SP, Brazil, marina.dias@bioetanol.org.br
# b School of Chemical Engineering, University of Campinas, PO Box 6066 – CEP
# 13083-970, Campinas – SP, Brazil
# LHV = 7565 # Bagasse lower heating value (kJ/kg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 28 17:28:09 2017

@author: Yoel Rene Cortes-Pena
"""
__all__ = ['Unit', 'PowerUtility', 'HeatUtility',
           'System', 'TEA', 'CombinedTEA', 
           'find', 'Flowsheet', 'CE']

from lazypkg import LazyPkg
#: Chemical engineering plant cost index (defaults to 567.5 at 2017)
CE = 567.5 

# %% Initialize BioSTEAM 

from ._heat_utility import HeatUtility
from ._power_utility import PowerUtility
from ._unit import Unit
from ._system import System
from ._tea import CombinedTEA, TEA
from ._flowsheet import Flowsheet, find


LazyPkg(__name__, ['utils', 'units', 'evaluation'])





# -*- coding: utf-8 -*-
"""
Created on Sun Dec  8 02:01:41 2019

@author: yoelr
"""

__all__ = ('test_biorefineries', 'test_lipidcane', 'test_cornstover')

def test_biorefineries():
    test_lipidcane()
    test_cornstover()

def test_lipidcane():
    from biorefineries.lipidcane import system
    IRR = system.lipidcane_tea.IRR
    assert 0.180 < IRR < 0.190, ("IRR of the lipid-cane biorefinery changed significantly "
                                f"from 0.185 to {IRR}")
    
def test_cornstover():
    from biorefineries.cornstover import system
    MESP = system.cornstover_tea.solve_price(system.ethanol)
    assert 0.70 < MESP < 0.74, ("MESP of the corn stover biorefinery changed significantly "
                               f"from 0.747 to {MESP} USD/kg")
    
if __name__ == '__main__':
    test_biorefineries()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 11:49:52 2018

This module includes classes and functions relating exception handling.

@author: Yoel Rene Cortes-Pena
"""
# import sys as _sys

__all__ = ('DesignError',)

# %% Biosteam errors

class DesignError(RuntimeError):
    """RuntimeError regarding unit design."""

# class UnitError(RuntimeError):
#     """RuntimeError regarding unit operations."""
#     def __init__(self, unit, method, error):
#         # Add location to message
#         msg = (f'{type(error).__name__} at {repr(unit)}.{method.__name__}\n{error}')
        
#         # Raise exception with same traceback but new message
#         super().__init__(msg)
#         self.original_exception = error
#         self.with_traceback(_sys.exc_info()[2])


#%% Decorators and functions

# def _try_method(method):
#     try: return method()
#     except UnitError:
#         raise UnitError
#     except Exception as error:
#         # If exception already include location, it is replaced
#         try:
#             self = method.__self__
#         except:
#             raise error
#         raise UnitError(self, method, error)
        




# -*- coding: utf-8 -*-
"""
Created on Sat Sep  1 17:35:28 2018

@author: yoelr
"""
import numpy as np
from ..utils.piping import StreamSequence
from .. import Unit, PowerUtility

__all__ = ('MassBalance',)

# %% Mass Balance Unit

class MassBalance(Unit):
    """Create a Unit object that changes net input flow rates to satisfy output flow rates. This calculation is based on mass balance equations for specified IDs. 

    Parameters
    ----------
    IDs : tuple[str]
        Chemicals that will be used to solve mass balance linear equations. The number of chemicals must be same as the number of input streams varied.
    streams : tuple[int]
        Indices of input streams that can vary in net flow rate
    is_exact=True : bool, optional
        True if exact flow rate solution is required for the specified IDs.
    balance='flow' : {'flow', 'composition'}, optional
          * 'flow': Satisfy output flow rates
          * 'composition': Satisfy net output molar composition

    Examples
    --------
    :doc:`notebooks/MassBalance Example`
    
    """
    line = 'Balance'
    results = None
    _has_cost = False
    _N_ins = 2
    heat_utilities = ()

    def __init__(self, ID='', outs=(), ins=None,
                 chemical_IDs=None, streams=None,
                 is_exact=True, balance='flow', thermo=None):
        self.ID = ID
        thermo = self._load_thermo(thermo)
        self._ins = StreamSequence(self._N_ins, ins, thermo, fixed_size=False)
        self._outs = StreamSequence(self._N_outs, outs, thermo, fixed_size=False)
        self._power_utility = PowerUtility()
        self.streams = streams
        self.chemical_IDs = chemical_IDs
        self.is_exact = is_exact
        self.balance = balance
        
    def _run(self):
        """Solve mass balance by iteration."""
        # SOLVING BY ITERATION TAKES 15 LOOPS FOR 2 STREAMS
        # SOLVING BY LEAST-SQUARES TAKES 40 LOOPS
        stream_index = self.streams
        balance = self.balance
        solver = np.linalg.solve if self.is_exact else np.linalg.lstsq
        ins = self.ins

        # Set up constant and variable streams
        vary = [ins[i] for i in stream_index]  # Streams to vary in mass balance
        constant = [i for i in ins if (i and i not in vary)]  # Constant streams
        index = self.chemicals.get_index(self.chemical_IDs)

        if balance == 'flow':
            # Perform the following calculation: Ax = b = f - g
            # Where:
            #    A = flow rate array
            #    x = factors
            #    b = target flow rates
            #    f = output flow rates
            #    g = constant input flow rates

            # Solve linear equations for mass balance
            A = np.array([s.mol for s in vary]).transpose()[index, :]
            f = self.mol_out[index]
            g = sum([s.mol[index] for s in constant])
            b = f - g
            x = solver(A, b)

            # Set flow rates for input streams
            for factor, s in zip(x, vary):
                s.mol[:] = s.mol * factor

        elif balance == 'fraction':
            # Perform the following calculation:
            # Ax = b
            #    = sum( A_ * x_guess + g_ )f - g
            #    = A_ * x_guess * f - O
            # O  = sum(g_)*f - g
            # Where:
            # A_ is flow array for all species
            # g_ is constant flows for all species
            # Same variable definitions as in 'flow'

            # Set all variables
            A_ = np.array([s.mol for s in vary]).transpose()
            A = np.array([s.mol for s in vary]).transpose()[index, :]
            f = self.z_mol_out[index]
            g_ = sum([s.mol for s in constant])
            g = g_[index]
            O = sum(g_) * f - g

            # Solve by iteration
            x_guess = np.ones_like(index)
            not_converged = True
            while not_converged:
                # Solve linear equations for mass balance
                b = (A_ * x_guess).sum()*f + O
                x_new = solver(A, b)
                not_converged = sum(((x_new - x_guess)/x_new)**2) > 0.0001
                x_guess = x_new

            # Set flow rates for input streams
            for factor, s in zip(x_new, vary):
                s.mol = s.mol * factor
        
        else:
            raise ValueError( "balance type must be one of the following: 'flow', 'composition'")


# %% Energy Balance Unit

# class EnergyBalance(Unit):
#     """Create a Unit object that changes a stream's temperature, flow rate, or vapor fraction to satisfy energy balance.

#     **Parameters**

#         **index:** [int] Index of stream that can vary in temperature, flow rate, or vapor fraction.
        
#         **Type:** [str] Should be one of the following
#             * 'T': Vary temperature of output stream
#             * 'F': Vary flow rate of input/output stream
#             * 'V': Vary vapor fraction of output stream
        
#         **Qin:** *[float]* Additional energy input.
        
#     .. Note:: This is not a mixer, input streams and output streams should match flow rates.

#     """
#     _kwargs = {'index': None,
#                'Type': 'T',
#                'Qin': 0}
#     line = 'Balance'
#     _has_cost = False
#     _graphics = MassBalance._graphics
#     _init_ins = MassBalance._init_ins
#     _init_outs = MassBalance._init_outs
    
#     def _run(self):        # Get arguments
#         ins = self.ins.copy()
#         outs = self.outs.copy()
#         kwargs = self._kwargs
#         index = kwargs['index']
#         Type = kwargs['Type']
#         Qin = kwargs['Qin']
        
#         # Pop out required streams
#         if Type == 'F':
#             s_in = ins.pop(index)
#             s_out = outs.pop(index)
#         else:
#             s = outs.pop(index)
        
#         # Find required enthalpy
#         H_in = sum(i.H for i in ins) + Qin
#         H_out = sum(o.H for o in outs)
#         H_s = H_out - H_in
        
#         # Set enthalpy
#         if Type == 'T':
#             s.H = -H_s
#         elif Type == 'V':
#             s.enable_phases()
#             s.VLE(Qin=s.H - H_s)
#         elif Type == 'F':
#             s.mol *= (s_out.H - s_in.H)/H_s
#         else:
#             raise ValueError(f"Type must be 'T', 'V' or 'F', not '{Type}'")
            
        
        


# -*- coding: utf-8 -*-
"""
Created on Mon Apr 29 18:28:55 2019

@author: yoelr
"""
from . import Facility
from ...utils.stream_link_options import static
from ..decorators import cost
from ... import HeatUtility
import numpy as np

__all__ = ('ChilledWaterPackage',)

@static
@cost('Duty', S=-14*4184000, kW=3400*0.7457, cost=1375e3, CE=551, n=0.7, BM=1.5)
class ChilledWaterPackage(Facility):
    """Create a chilled water package that is cost based on flow rate of chilled water.
    
    Parameters
    ----------
    Duty : float
        Chilled water duty (kJ/hr)
    
    """
    _N_heat_utilities = 1
    _units = {'Duty': 'kJ/hr'}
    def __init__(self, ID=''):
        thermo = HeatUtility.cooling_agents['Chilled water'].thermo
        super().__init__(ID, 'return_chilled_water', 'chilled_water', thermo=thermo)
        self.chilled_water_utilities = set()
        
    def _design(self):
        cwu = self.chilled_water_utilities
        if not cwu:
            for u in self.system.units:
                if u is self: continue
                for hu in u.heat_utilities:
                    if hu.ID == 'Chilled water': cwu.add(hu)
        self.design_results['Duty'] = duty = sum([i.duty for i in cwu])
        hu = self.heat_utilities[0]
        hu(duty, 330)
        used = self._ins[0]
        used.mol[0] = sum([i.flow for i in cwu])
        used.T = np.array([i.outlet_utility_stream.T for i in cwu]).mean()
        self.outs[0].T = hu.cooling_agents['Chilled water'].T
        


# -*- coding: utf-8 -*-
"""
Created on Sun Dec  8 02:13:07 2019

@author: yoelr
"""
import biosteam as bst
import doctest

__all__ = ('test_binary_distillation', 'test_mixer', 'test_splitter')

def test_binary_distillation():
    doctest.testmod(bst.units._distillation)
    
def test_mixer():
    doctest.testmod(bst.units._mixer)

def test_splitter():
    doctest.testmod(bst.units._splitter)
    
def test_unit_operations():
    test_binary_distillation()
    test_mixer()
    test_splitter()
    
if __name__ == '__main__':
    test_unit_operations()


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 10:57:12 2019

@author: yoelr
"""
from .._unit import Unit
from .decorators import cost

@cost('Flow rate', units='kg/hr', cost=2.5e6,
      CE=567.3, n=0.6, S=500e3, kW=3000, BM=1.39)
class Shredder(Unit):  pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 11:49:52 2018

@author: Yoel Rene Cortes-Pena
"""
from . import array_utils
from . import color_utils
from . import other_utils
from . import stream_utils
from . import plot_utils
from . import tictoc
from . import not_implemented_method
from . import network_utils
from . import piping
from . import stream_link_options
from . import design_warning

__all__ = ('array_utils', 
           'color_utils',
           'other_utils', 
           'stream_utils',
           'plot_utils', 
           'tictoc',
           'not_implemented_method',
           'network_utils',
           *plot_utils.__all__, 
           *not_implemented_method.__all__, 
           *array_utils.__all__, 
           *color_utils.__all__,
           *other_utils.__all__, 
           *stream_utils.__all__, 
           *tictoc.__all__, 
           *network_utils.__all__, 
           *piping.__all__, 
           *stream_link_options.__all__,
           *design_warning.__all__)

from .not_implemented_method import *
from .array_utils import *
from .color_utils import *
from .other_utils import *
from .stream_utils import *
from .plot_utils import *
from .tictoc import *
from .network_utils import *
from .piping import *
from .stream_link_options import *
from .design_warning import *



# -*- coding: utf-8 -*-
"""
Created on Thu May  9 13:38:57 2019

@author: Guest Group
"""
from ._block import Block
from .. import Unit
from thermosteam import Stream
import numpy as np
import pandas as pd
from chaospy import J
from .evaluation_tools import load_default_parameters

__all__ = ('State',)

# %% functions

def param_unit(param):
    element = param.element
    if isinstance(element, Unit): return element
    elif isinstance(element, Stream): return element._sink

def parameter(system, element, setter, kind, name, distribution, units, baseline):
    if kind == 'coupled':
        return Block(element, system).parameter(setter, name=name,
                    distribution=distribution, units=units, baseline=baseline)
    elif kind == 'isolated':
        return Block(element, None).parameter(setter, name=name,
                    distribution=distribution, units=units, baseline=baseline)
    elif kind == 'design':
        return Block(element, None).parameter(setter, element._summary,
                    name, distribution=distribution, units=units, baseline=baseline)
    elif kind == 'cost':
        if hasattr(element, '_end'):
            def simulate():
                element._cost()
                element._end()
        else:
            simulate = element._cost
        return Block(element, None).parameter(setter, simulate, name,
                    distribution=distribution, units=units, baseline=baseline)
    raise ValueError(f"kind must be either 'coupled', 'isolated', 'design', or 'cost' (not {kind}).")

class UpdateWithSkipping:
    __slots__ = ('cache', 'params')
    def __init__(self, params):
        self.params = tuple(params)
        self.cache = None
    def __call__(self, sample):
        try:
            sim = None
            for p, x, same in zip(self.params, sample, self.cache==sample):
                if same: continue
                p.setter(x)
                if sim: continue
                if p.system: sim = p.simulate 
                else: p.simulate()
            if sim: sim()
            self.cache = sample.copy()
        except Exception as Error:
            self.cache = None
            raise Error

class UpdateWithoutSkipping:
    __slots__ = ('cache', 'params')
    def __init__(self, params):
        self.params = tuple(params)
        self.cache = None
    def __call__(self, sample):
        sim = None
        for p, x in zip(self.params, sample):
            p.setter(x)
            if sim: continue
            if p.system: sim = p.simulate 
            else: p.simulate()
        if sim: sim()



# %%
    
class State:
    """Create a State object that can update the `system` state given a sample of parameter states.
    
    Parameters
    ----------
    system : System
    skip=False : bool
        If True, skip simulation for repeated states
    
    """
    __slots__ = ('_system', # [System]
                 '_params', # list[Parameter] All parameters.
                 '_update', # [function] Updates system state.
                 '_skip') # [bool] If True, skip simulation for repeated states
    
    load_default_parameters = load_default_parameters
    
    def __init__(self, system, skip=False):
        self._system = system
        self._params = []
        self._update = None
        self._skip = skip
    
    def __len__(self):
        return len(self._params)
    
    def copy(self):
        """Return copy."""
        copy = self.__new__(type(self))
        copy._params = list(self._params)
        copy._system = self._system
        copy._update = self._update
        copy._skip = self._skip
        return copy
    
    def get_parameters(self):
        """Return parameters."""
        if not self._update: self._loadparams()
        return tuple(self._params)
    
    def get_distribution_summary(self):
        """Return dictionary of shape name-DataFrame pairs."""
        params = self.get_parameters()
        if not params: return None
        params_by_shape = {}
        shape_keys = {}
        for p in params:
            distribution = p.distribution
            if not distribution: continue
            shape = type(distribution).__name__
            if shape in params_by_shape:
                params_by_shape[shape].append(p)
            else:
                params_by_shape[shape] = [p]
                shape_keys[shape] = tuple(distribution._repr.keys())
        tables_by_shape = {}
        for shape, params in params_by_shape.items():
            data = []
            columns = ('Element', 'Name', 'Units', 'Shape', *shape_keys[shape])
            for p in params:
                distribution = p.distribution
                element = p.element_name
                name = p.name.replace(element, '')
                units = p.units or ''
                values = distribution._repr.values()
                data.append((element, name, units, shape, *values))
            tables_by_shape[shape] =  pd.DataFrame(data, columns=columns)
        return tables_by_shape
    
    def parameter(self, setter=None, element=None, kind='isolated',
                  name=None, distribution=None, units=None, baseline=None):
        """Define and register parameter.
        
        Parameters
        ----------    
        setter : function
                 Should set parameter in the element.
        element : Unit or Stream
                  Element in the system being altered.
        kind : {'coupled', 'isolated', 'design', 'cost'}
            * 'coupled': parameter is coupled to the system.
            * 'isolated': parameter does not affect the system but does affect the element (if any).
            * 'design': parameter only affects design and cost of the element.
            * 'cost': parameter only affects cost of the element.
        name : str
               Name of parameter. If None, default to argument name of setter.
        distribution : chaospy.Dist
                       Parameter distribution.
        units : str
                Parameter units of measure
        baseline : float
            Baseline value of parameter.
        
        Notes
        -----
        
        If kind is 'coupled', account for downstream operations. Otherwise, only account for given element. If kind is 'design' or 'cost', element must be a Unit object.
        
        """
        if not setter:
            return lambda setter: self.parameter(setter, element, kind, name,
                                                 distribution, units, baseline)
        param = parameter(self._system, element, setter, kind,
                          name, distribution, units, baseline)
        self._params.append(param)
        self._erase()
        return setter
    
    def sample(self, N, rule): 
        """Return N samples from parameter distribution at given rule.
        
        Parameters
        ----------
        N : int
            Number of samples.
        rule : str
               Sampling rule.
        
        Notes
        -----
        Use the following ``rule`` flag for sampling:
        
        +-------+-------------------------------------------------+
        | key   | Description                                     |
        +=======+=================================================+
        | ``C`` | Roots of the first order Chebyshev polynomials. |
        +-------+-------------------------------------------------+
        | ``NC``| Chebyshev nodes adjusted to ensure nested.      |
        +-------+-------------------------------------------------+
        | ``K`` | Korobov lattice.                                |
        +-------+-------------------------------------------------+
        | ``R`` | Classical (Pseudo-)Random samples.              |
        +-------+-------------------------------------------------+
        | ``RG``| Regular spaced grid.                            |
        +-------+-------------------------------------------------+
        | ``NG``| Nested regular spaced grid.                     |
        +-------+-------------------------------------------------+
        | ``L`` | Latin hypercube samples.                        |
        +-------+-------------------------------------------------+
        | ``S`` | Sobol low-discrepancy sequence.                 |
        +-------+-------------------------------------------------+
        | ``H`` | Halton low-discrepancy sequence.                |
        +-------+-------------------------------------------------+
        | ``M`` | Hammersley low-discrepancy sequence.            |
        +-------+-------------------------------------------------+
            
        """
        if not self._update: self._loadparams()
        return J(*[i.distribution for i in self._params]).sample(N, rule).transpose()
    
    def _erase(self):
        """Erase cached data."""
        self._update = None
    
    def _loadparams(self):
        """Load parameters."""
        length = len(self._system._unit_network)
        index =  self._system._unit_network.index
        self._params.sort(key=lambda x: index(param_unit(x)) if x.system else length)
        self._update = (UpdateWithSkipping if self._skip 
                        else UpdateWithoutSkipping)(self._params)
    
    def __call__(self, sample):
        """Update state given sample of parameters."""
        if not self._update: self._loadparams()
        return self._update(np.asarray(sample, dtype=float))
    
    def _repr(self):
        return f'{type(self).__name__}: {self._system}'
    
    def __repr__(self):
        return '<' + self._repr() + '>'
    
    def _info(self):
        if not self._update: self._loadparams()
        if not self._params: return f'{self._repr()}\n (No parameters)'
        lines = []
        lenghts_block = []
        lastblk = None
        for i in self._params:
            blk = i.element_name
            element = len(blk)*' ' if blk==lastblk else blk
            lines.append(f" {element}${i.name}\n")
            lastblk = blk
            lenghts_block.append(len(blk))
        maxlen_block = max(lenghts_block)
        out = f'{self._repr()}\n'
        maxlen_block = max(maxlen_block, 7)
        out += ' Element:' + (maxlen_block - 7)*' ' + ' Parameter:\n'
        for newline, len_ in zip(lines, lenghts_block):
            newline = newline.replace('$', ' '*(maxlen_block-len_) + '  ')
            out += newline
        return out.rstrip('\n ')
    
    def show(self):
        """Return information on metric parameters."""
        print(self._info())
    _ipython_display_ = show
    


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 17:19:33 2018

This module includes classes and functions relating string coloring and format.

@author: Guest Group
"""
from colorpalette import Color, Palette
import numpy as np

__all__ = ('colors',)

# %% Classes for coloring

# BioSTEAM console
ansicolor = Color.from_ansi
colors = Palette()
colors.dim = ansicolor('dim', '\x1b[37m\x1b[22m')
colors.reset = ansicolor('reset', '\x1b[0m')
colors.exception = ansicolor('bright red', '\x1b[31m\x1b[1m')
colors.next = colors.info = ansicolor('bright blue', '\x1b[34m\x1b[1m')
colors.note = ansicolor('bright cyan', '\x1b[36m\x1b[1m')
colors.dim = ansicolor('dim', '\x1b[37m\x1b[22m')

# Main colors
colors.blue = Color('blue', np.array([0.376, 0.757, 0.812]) * 255)
colors.blue_tint = colors.blue.tint(50)
colors.blue_shade = Color('special blue shade', [53, 118, 127])
colors.blue_dark = colors.blue.shade(70)
colors.orange = Color('orange', 255*np.array([0.976, 0.561, 0.376]))
colors.orange_tint = colors.orange.tint(50)
colors.orange_shade = colors.orange.shade(50)
colors.red = Color('red', (241, 119, 127))
colors.red_tint = colors.red.tint(50)
colors.red_shade = colors.red.shade(25)
colors.red_dark = colors.red.shade(45)
colors.green = Color('green', '#79bf82')
colors.green_tint = colors.green.tint(50)
colors.green_shade = colors.green.shade(25)
colors.green_dark = colors.green.shade(45)
colors.yellow = Color('yellow', '#f3c354')
colors.yellow_tint = colors.yellow.tint(45)
colors.yellow_shade = colors.yellow.shade(35)
colors.yellow_dark = colors.yellow.shade(55)
                     
# Contrast colors
colors.green = Color('green', '#33CC33')
colors.green_tint =colors. green.tint(20).shade(15)
colors.green_shade = colors.green.shade(50).tint(40).shade(30)
colors.purple = Color('purple', 255*np.array([0.635, 0.502, 0.725]))
colors.purple_tint = colors.purple.tint(60)
colors.purple_shade = colors.purple.shade(25)
colors.purple_dark = colors.purple.shade(75)

# Neutral colors
colors.brown = Color('brown', [97, 30, 9])
colors.brown_tint = colors.brown.tint(60)
colors.brown_shade = colors.brown.shade(25)
colors.grey = Color('grey', 255*np.array([0.565, 0.569, 0.557]))
colors.grey_tint = colors.grey.tint(25)
colors.grey_shade = colors.grey.shade(25)
colors.neutral = Color('neutral', 255*np.array([0.4824, 0.5216, 0.502]))
colors.neutral_tint = colors.neutral.tint(50)
colors.neutral_shade = colors.neutral.shade(50)



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 15:47:26 2018

@author: yoelr
"""

from .._unit import Unit
from .._exceptions import DesignError
from ._mixer import Mixer
from .decorators import cost
import numpy as np
import biosteam as bst

class Tank(Unit, isabstract=True):
    """Abstract Tank class."""
    _units = {'Total volume': 'm^3'}
    _N_outs = 1
    
    @property
    def tau(self):
        """Residence time (hr)."""
        return self._tau
    @tau.setter
    def tau(self, tau):
        self._tau = tau


class StorageTank(Tank):
    r"""Create a storage tank with volume based on residence time [1].

    .. math::

        V &= \\tau Q \\\\
        C_{fob}^{2007} &= 250000 + 94.2 V (2*10^3 < V < 50*10^3 m^3) \\\\

    References
    ----------
    [1] Apostolakou, A. A., Kookos, I. K., Marazioti, C., & Angelopoulos, K. C. (2009). Techno-economic analysis of a biodiesel production process from vegetable oils. Fuel Processing Technology, 90(7–8), 1023–1031. https://doi.org/10.1016/j.fuproc.2009.04.017

    """
    line = 'Storage tank'
    
    #: [float] residence time (hr)
    _tau = 4*7*24

    def _get_design_specs(self):
        return (('Residence time', self.tau/24, 'days'),)

    def _design(self):
        V = self._tau * self.F_vol_out
        self.design_results['Number of tanks'] = N = np.ceil(V/50e3)
        self.design_results['Total volume'] = V/N

    def _cost(self):
        N = self.design_results['Number of tanks']
        V = self.design_results['Total volume']
        if V < 2e3:
            self.purchase_costs['Tank'] = N * (65000 + 158.7*(V/N)) * bst.CE/525.4
        elif V < 50e3:
            self.purchase_costs['Tank'] = N * (250000 + 94.2*(V/N)) * bst.CE/525.4
        else:    
            raise DesignError(f"Volume is out of bounds for costing")

@cost('Total volume', cost=12080, CE=525.4, n=0.525, kW=0.0985, ub=30,
      N='Number of tanks')
class MixTank(Tank):
    """Create a mixing tank with volume based on residence time.

    .. math::

       V &= \\frac{\\tau* Q}{0.8}

       C_{fob}^{2007} &= 12080 * V^{0.525} (0.1 < V < 30 m^3)

    References
    ----------
    [1] I.K. Kookos, Preliminary Chemical Process Synthesis and Design, Tziolas Publishing, Thessalonika, Greece, 2008 (book in Greek).

    """
    line = 'Mix tank'
    _tau = 1
    _N_ins = 2
    _run = Mixer._run
    _minimum_volume = 0.1
    
    #: Fraction of working volume
    _V_wf = 0.8
    
    def _get_design_specs(self):
        return (('Residence time', self.tau, 'hr'),
                ('Working volume fraction', self.working_volume_fraction, ''))
    
    @property
    def working_volume_fraction(self):
        """Fraction of working volume."""
        return self._V_wf

    @working_volume_fraction.setter
    def working_volume_fraction(self, V_wf):
        if not 0 < V_wf <= 1:
            raise ValueError('working_volume_fraction must be between 0 to 1')
        self._V_wf = V_wf
    
    def _design(self):
        self.design_results['Total volume'] = self._tau * self.outs[0].F_vol / self._V_wf
        # if V < self._minimum_volume: self._lb_warning('Volume', V, 'm^3', self._minimum_volume)




# -*- coding: utf-8 -*-
"""
Created on Tue Sep 10 09:01:47 2019

@author: yoelr
"""
import biosteam as bst
from .other_utils import strtuple

__all__ = ('_add_upstream_neighbors', '_add_downstream_neighbors',
           'Thread', 'build_network')

def _add_upstream_neighbors(unit, set):
    """Add upsteam neighboring units to set."""
    for s in unit._ins:
        u_source = s._source
        if u_source: set.add(u_source)

def _add_downstream_neighbors(unit, set):
    """Add downstream neighboring units to set."""
    for s in unit._outs:
        u_sink = s._sink
        if u_sink: set.add(u_sink)

def build_network(sources, sinks=None):
    thread = Thread.from_units(sources, sinks)
    return thread.build_network()

def traverse(iterable):
    L = list
    inst = isinstance
    path = []
    for iter in iterable:
        if inst(iter, L):
            path += traverse(iter)
        else:
            path.append(iter)
    return path

def is_ascending(nested_index, other):
    for depth, jk in enumerate(zip(nested_index, other)): 
        j, k = jk
        if j != k: break
    return depth, j < k

def get_nested_index(iterable, item):
    L = list
    inst = isinstance
    index = []
    for i, iter in enumerate(iterable):
        if inst(iter, L):
            index += get_nested_index(iter, item)
        else:
            if iter is item:
                index.append(i)
                break
    return index

def get_nested_item(iterable, nested_index):
    for i in nested_index:
        iterable = iterable[i]
    return iterable

class Thread:
    __slots__ = ('path', 'source', 'sink')
    
    def __init__(self, path, source, sink):
        #: list[Unit or ...] Ordered units in the thread.
        self.path = path
        
        #: [Unit or None] Origin of thread.
        self.source = source
        
        #: [Unit or None] End of thread
        self.sink = sink
        
    @classmethod
    def single_thread(cls, start, source, sinks, thread_args, past_sources):
        path = [start]
        next_unit = start
        sinks = set(sinks)
        neighbors = set()
        while next_unit:
            next_units = [i.sink for i in next_unit.outs if i.sink]
            if next_units:
                last_unit = next_unit
                next_unit, *other_units = next_units
                while True:
                    new_neighbors = [i for i in other_units if i not in neighbors]
                    if not new_neighbors: break
                    neighbors.update(new_neighbors)
                    if next_unit not in past_sources:
                        past_sources.add(next_unit)
                        thread_args.append((next_unit, last_unit))
                    next_unit, *other_units = new_neighbors
                if next_unit in sinks:
                    break
                elif next_unit in path:
                    index = path.index(next_unit)
                    recycle_units = path[index:]
                    path = path[:index]
                    path.append(recycle_units)
                    break
                path.append(next_unit)
            else:
                break
        self = cls(path, source, sink=next_unit)
        sinks.update(traverse(path))
        return self
        
    @classmethod
    def web(cls, start, source, sinks, thread_args, threads, past_sources):
        more = []
        threads.extend([cls.single_thread(*i, sinks, more, past_sources) for i in thread_args])
        if more:
            cls.web(start, source, sinks, more, threads, past_sources)
        return threads
        
    @classmethod
    def from_unit(cls, start, source, sinks):
        thread_args = []
        past_sources = set()
        self = cls.single_thread(start, source, sinks, thread_args, past_sources)
        threads = []
        cls.web(start, source, sinks, thread_args, threads, past_sources)
        for thread in threads: self.knit(thread)
        return self

    @classmethod
    def from_units(cls, sources, sinks):
        self, *threads = [cls.from_unit(i, None, sinks) for i in sources]
        for thread in threads: self.knit(thread)
    
    def knit(self, other):
        source = other.source
        sink = other.sink
        path = self.path
        source_index = get_nested_index(path, source)
        sink_index = get_nested_index(path, sink)
        self.show()
        other.show()
        if not sink_index:
            self.path.extend(other.path)
            self.show()
            return
        depth, check = is_ascending(source_index, sink_index)
        if check:
            *nest_index, item_index = sink_index
            sink_nest = get_nested_item(path, nest_index)
            sink_nest[:] = sink_nest[:item_index] + other.path + sink_nest[item_index:]
        else:
            *nest_index, sink_item_index = sink_index[:depth+1]
            source_item_index = source_index[depth]
            nest = get_nested_item(path, nest_index)
            stop = source_item_index+1
            segment = slice(sink_item_index, stop)
            nest[segment] = [[nest[segment] + [other.path]]] + nest[stop:]
        self.show()
    
    def build_network(self):
        units = self.path + [self.sink]
        network = []
        for segment in units:
            if isinstance(segment, list):
                source = segment[0]
                for recycle in segment[-1].outs:
                    if recycle.sink is source: break
                system = bst.System(segment, recycle=recycle)
                network.append(system)
            else:
                network.append(segment)
        return network

    def show(self):
        path = strtuple(self.path)
        i = 1; last_i = 0
        while True:
            i += 2
            i = path.find(', ', i)
            i_next = path.find(', ', i+2)
            if (i_next-last_i) > 35:
                path = (path[:i] + '%' + path[i:])
                last_i = i
            elif i == -1: break
        path = path.replace('%, ', ',\n'+' '*8)
        print(f"{type(self).__name__}:\n"
             + f" source: {self.source}\n"
             + f" sink: {self.sink}\n"
             + f" path: {path}")

# Example path
# >>> threads = [
# ...     Thread(units=[U0, U1, [U2, U3, U4, U5]],   source=None,   sink=None),
# ...     Thread(units=[U6, U7],                     source=U3,     sink=U5),
# ...     Thread(units=[U8, U9, U0],                 source=U6,     sink=U0),
# ... ]
# >>> path = Thread.path(threads)
# >>> path
# [System(path=[U0, U1, System(path=[U2, U3, U4, U6, U7, U5], recycle=stream), U8, U9])]

# First knit threads toghether

# step 1:
# [Thread(units=[U0, U1, [U2, U3, U4, U6, U7, U5]], source=None, sink=None),
#  Thread(units=[U8, U9, U0],                       source=U3,   sink=U0]
# step 2:
# [Thread(units=[[U0, U1, [U2, U3, U4, U6, U7, U5], U8, U9]], source=None, sink=None)]

# Create systems and subsystems
# [System(path=[U0, U1, System(path=[U2, U3, U4, U6, U7, U5], recycle=stream), U8, U9])]




# -*- coding: utf-8 -*-
"""
Created on Mon May 20 22:04:02 2019

@author: yoelr
"""
from .. import Unit
import numpy as np 
from thermosteam.indexer import ChemicalIndexer
__all__ = ('run_split', 'run_split_with_mixing')


def run_split(self):
    """Splitter mass and energy balance function based on one input stream."""
    top, bot = self.outs
    feed = self._ins[0]
    top_mol = top.mol
    feed_mol = feed.mol
    top_mol[:] = feed_mol * self.split
    bot.mol[:] = feed_mol - top_mol
    bot.T = top.T = feed.T
    bot.P = top.P = feed.P

def run_split_with_mixing(self):
    """Splitter mass and energy balance function with mixing all input streams."""
    top, bot = self.outs
    ins = self.ins
    top.mix_from(ins)
    bot.copy_like(top)
    top_mol = top.mol
    top_mol[:] *= self.split
    bot.mol[:] -= top_mol


class Splitter(Unit):
    """Create a splitter that separates mixed streams based on splits.

    Parameters
    ----------
    ins
        [0] Feed stream
    outs
        [0] Split stream
        
        [1] Remainder stream    
    split : Should be one of the following
            * [float] The fraction of net feed in the 0th output stream
            * [array_like] Componentwise split of feed to 0th output stream
    order=None : Iterable[str], defaults to Stream.species.IDs
        Species order of split.
    
    Examples
    --------
    Create a Splitter object with an ID, a feed stream, two output streams,
    and an overall split:
        
    .. code-block:: python
    
       >>> from biosteam import units
       >>> import thermosteam as tmo
       >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'])
       >>> tmo.settings.set_thermo(chemicals)
       >>> feed = tmo.Stream('feed', Water=20, Ethanol=10, T=340)
       >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'), split=0.1)
       >>> S1.simulate()
       >>> S1.show()
       Splitter: S1
       ins...
       [0] feed
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    20
                           Ethanol  10
       outs...
       [0] top
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    2
                           Ethanol  1
       [1] bot
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    18
                           Ethanol  9
      
    Create a Splitter object, but this time with a componentwise split:
        
    .. code-block:: python
    
       >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
       ...                     split={'Water': 0.1, 'Ethanol': 0.99})
       >>> S1.simulate()
       >>> S1.show()
       Splitter: S1
       ins...
       [0] feed
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    20
                           Ethanol  10
       outs...
       [0] top
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    2
                           Ethanol  9.9
       [1] bot
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    18
                           Ethanol  0.1
                           
    Create a Splitter object using componentwise split, but this time specify the order:
        
    .. code-block:: python
    
       >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
       ...                     order=('Ethanol', 'Water'),
       ...                     split=(0.99, 0.10))
       >>> S1.simulate()
       >>> S1.show()
       Splitter: S1
       ins...
       [0] feed
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    20
                           Ethanol  10
       outs...
       [0] top
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    2
                           Ethanol  9.9
       [1] bot
           phase: 'l', T: 340 K, P: 101325 Pa
           flow (kmol/hr): Water    18
                           Ethanol  0.1

    """
    _N_outs = 6
    
    @property
    def isplit(self):
        """[ChemicalIndexer] Componentwise split of feed to 0th output stream."""
        return self._isplit
    @property
    def split(self):
        """[Array] Componentwise split of feed to 0th output stream."""
        return self._isplit._data
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, split, order=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        chemicals = self.thermo.chemicals
        if isinstance(split, dict):
            assert not order, "cannot pass 'other' key word argument when split is a dictionary"
            order, split = zip(*split.items())
        
        if order:
            isplit = chemicals.iarray(order, split)
        elif hasattr(split, '__len__'):
            isplit = ChemicalIndexer.from_data(np.asarray(split),
                                               phase=None,
                                               chemicals=chemicals)
        else:
            split = split * np.ones(chemicals.size)
            isplit = ChemicalIndexer.from_data(split,
                                               phase=None,
                                               chemicals=chemicals)
        self._isplit = isplit

    def _run(self):
        top, bot = self.outs
        feed, = self.ins
        feed_mol = feed.mol
        bot.T = top.T = feed.T
        bot.P = top.P = feed.P
        bot.phase = top.phase = feed.phase
        top.mol[:] = top_mol = feed_mol * self.split
        bot.mol[:] = feed_mol - top_mol

Splitter._N_outs = 2

class InvSplitter(Unit):
    """Create a splitter that sets the input stream based on output streams. Must have only one input stream. The output streams will become the same temperature, pressure and phase as the input.
    """
    _graphics = Splitter._graphics
    def _run(self):
        feed = self.ins[0]
        feed.mol[:] = self.mol_out
        T = feed.T
        P = feed.P
        phase = feed.phase
        for out in self._outs:
            out.T = T
            out.P = P
            out.phase = phase 
InvSplitter.line = 'Splitter'


        
    


# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:25:34 2018

@author: yoelr
"""
from thermosteam._stream import Stream, mol_units, mass_units, vol_units
from thermosteam.base.units_of_measure import convert, get_dimensionality, DisplayUnits
from thermosteam.exceptions import DimensionError
from thermosteam import Thermo

__all__ = ('HeatUtility',)

# Costs from Table 17.1 in Warren D.  Seider et al.-Product and Process Design Principles_ Synthesis, Analysis and Evaluation-Wiley (2016)
# ^This table was made using data from Busche, 1995
# Entry temperature conditions of coolants taken from Table 12.1 in Warren, 2016

# %% Default Heat Transfer chemicals

_Water = Thermo(['Water'])

# %% Utility classes

class UtilityAgent:
    """Create a UtilityAgent object that defines a utility option.
    
    Parameters
    ----------
    thermo: Thermo
    z_mol: tuple[float]
            Molar fraction of chemicals.
    phase : {'g', 'l'}
            Either gas ("g") or liquid ("l").
    Hvap : float
           Latent heat of vaporization (kJ/kmol)
    T_limit : float
              Temperature limit (K)
    price_kJ : float
               Price (USD/kJ)
    price_kmol : float
                 Price (USD/kmol)
    efficiency : float
                 Heat transfer efficiency (between 0 to 1)
            
    """
    __slots__ = ('thermo', 'z_mol', 'T', 'P', 'phase', 'Hvap', 'T_limit',
                 'price_kJ', 'price_kmol', 'efficiency')
    def __init__(self, thermo, z_mol, T, P, phase, Hvap, T_limit,
                 price_kJ, price_kmol, efficiency):
        self.thermo = thermo
        self.z_mol = z_mol
        self.T = T
        self.P = P
        self.phase = phase
        self.Hvap = Hvap
        self.T_limit = T_limit
        self.price_kJ = price_kJ
        self.price_kmol = price_kmol
        self.efficiency = efficiency
    
    def __repr__(self):
        return f"<{type(self).__name__}: T={self.T} K>"
    
    def _ipython_display_(self):
        print(f"{type(self).__name__}:\n"
             +f" thermo                Thermo({self.thermo.chemicals}, ...)\n"
             +f" z_mol                 ({', '.join([format(i,'.2f') for i in self.z_mol])},)\n"
             +f" T           K         {self.T:,.2f}\n"
             +f" P           Pa        {self.P:,.0f}\n"
             +f" phase                 '{self.phase}'\n"
             +f" Hvap        kJ/kmol   {self.Hvap and format(self.Hvap, ',.4g')}\n"
             +f" T_limit     K         {self.T_limit and format(self.T_limit, ',.2f')}\n"
             +f" price_kJ    USD/kJ    {self.price_kJ:.4g}\n"
             +f" price_kmol  USD/kmol  {self.price_kmol:.4g}\n"
             +f" efficiency            {self.efficiency:.2f}")
        
    show = _ipython_display_

            
class UtilityAgents:
    __slots__ = ('_agents',)
    
    def __init__(self, agents):
        self._agents = dict(agents)
        
    def __getitem__(self, key):
        return self._agents[key]
    
    def __setitem__(self, ID, agent):
        assert isinstance(agent, UtilityAgent)
        agents = sorted([*self._agents.items(), (ID, agent)], key=self._sorting_key)
        self._agents = dict(agents)
    
    def __iter__(self):
        yield from self._agents.items()
        
    def __repr__(self):
        return f"{type(self).__name__}({self._agents})"


class HeatingAgents(UtilityAgents):
    __slots__ = ()
    
    @staticmethod
    def _sorting_key(ID_agent):
        _, agent = ID_agent
        return agent.T
    
    def select_agent(self, T_pinch):
        """Return a heating agent that works at the pinch temperature and return UtilityAgent.
        
        Parameters
        ----------
        T_pinch : float
                  Pinch temperature of process stream (K)
        
        """
        for ID, agent in self:
            if T_pinch < agent.T: return ID, agent
        raise RuntimeError(f'no heating agent that can heat over {T_pinch} K')
    

class CoolingAgents(UtilityAgents):
    __slots__ = ()
    
    @staticmethod
    def _sorting_key(ID_agent):
        _, agent = ID_agent
        return 1 / agent.T

    # Selection of a heat transfer agent
    def select_agent(self, T_pinch):
        """Select a cooling agent that works at the pinch temperature and return UtilityAgent.
        
        Parameters
        ----------
        T_pinch : float
                  Pinch temperature of process stream (K)
        
        """
        for ID, agent in self:
            if T_pinch > agent.T: return ID, agent
        raise RuntimeError(f'no cooling agent that can cool under {T_pinch} K')


class HeatUtility:
    """Create an HeatUtility object that can choose a utility stream and calculate utility requirements. It can calculate required flow rate, temperature change, or phase change of utility. Calculations assume counter current flow rate.
    
    Parameters
    ----------
    efficiency=None : float, optional
        Enforced fraction of heat transfered from utility (due
        to losses to environment).
    
    Attributes
    ----------
    cooling_agents : dict[str: UtilityAgent]
                     All cooling utilities available
    heating_agents : dict[str: UtilityAgent]
                     All heating utilities available
    
    Examples
    --------
    Create a heat utility:
        
    .. code-block:: python
    
       >>> from biosteam import HeatUtility
       >>> hu = HeatUtility()
       >>> hu.show()
       HeatUtility: None
        duty: 0
        flow: 0
        cost: 0
    
    Calculate utility requirement by calling it with a duty (kJ/hr), and entrance and exit temperature (K):
        
    .. code-block:: python
    
       >>> hu(1000, 300, 350)
       >>> hu.show()
       HeatUtility: Low pressure steam
        duty: 1.05e+03 kJ/hr
        flow: 0.0271 kmol/hr
        cost: 0.00644 USD/hr
   
    All results are accessible:
        
    .. code-block:: python
    
       >>> hu.ID, hu.duty, hu.flow, hu.cost
       ('Low pressure steam', 1052.6315789473686, 0.027090580063500323, 0.006442139939100377)
           
    """
    __slots__ = ('inlet_utility_stream', 'outlet_utility_stream', 'ID', 'duty',
                 'flow', 'cost', 'efficiency', '_args')
    dT = 5  #: [float] Pinch temperature difference
    
    #: [DisplayUnits] Units of measure for IPython display
    display_units = DisplayUnits(duty='kJ/hr', flow='kmol/hr', cost='USD/hr')

    cooling_agents = CoolingAgents({
        'Cooling water': UtilityAgent(_Water, (1,), 305.372, 101325, 'l',
                                      None, 324.817, 0, 4.8785e-4, 1),
        'Chilled water': UtilityAgent(_Water, (1,), 280.372, 101325, 'l',
                                      None, 300.372, -5e-6, 0, 1),
        'Chilled Brine': UtilityAgent(_Water, (1,), 255.372, 101325, 'l',
                                      None, 275.372, -8.145e-6, 0, 1)})

    heating_agents = HeatingAgents({
        'Low pressure steam': UtilityAgent(_Water, (1,), 411.494, 344738.0, 'g',
                                           38856., None, 0, 0.2378, 0.95),
        'Medium pressure steam': UtilityAgent(_Water, (1,), 454.484, 1034214.0,
                                              'g', 36264., None, 0, 0.2756, 0.90),
        'High pressure steam': UtilityAgent(_Water, (1,), 508.858, 3102642.0, 'g',
                                            32124., None, 0, 0.3171, 0.85)})

    def __init__(self, efficiency=None):
        self.efficiency = efficiency
        self.empty()

    def _init_streams(self, flow, thermo, T, P, phase):
        """Initialize utility streams."""
        self.inlet_utility_stream = Stream(None, flow, thermo=thermo, T=T, P=P, phase=phase)
        self.outlet_utility_stream = self.inlet_utility_stream.flow_proxy()

    def empty(self):
        self.ID = ''
        self.cost = self.flow = self.duty = 0
        self._args = None
        #: tuple[bool, float] Cached arguments:
        #:     * [0]: [bool] True if duty is negative 
        #:     * [1]: [float] Pinch temperature of fresh utility
            

    def __call__(self, duty, T_in, T_out=None):
        """Calculate utility requirements given the essential parameters.
        
        Parameters
        ----------
        duty : float
               Unit duty requirement (kJ/hr)
        T_in : float
               Entering process stream temperature (K)
        T_out : float, optional
                Exit process stream temperature (K)
        
        """
        if duty == 0:
            self.empty()
            return
        T_out = T_out or T_in
        iscooling = duty < 0
        T_pinch_in, T_pinch_out = self._get_pinch(iscooling, T_in, T_out, self.dT)

        ## Select heat transfer agent ##
        
        args = (iscooling, T_pinch_in)
        agents = self.cooling_agents if iscooling else self.heating_agents
        if self._args == args:            
            # Use cached agent
            agent = agents[self.ID]
        else:
            # Select agent
            self._args = args
            ID, agent = agents.select_agent(T_pinch_in)
            self.load_agent(ID, agent)
            
        ## Calculate utility requirement ##
        
        efficiency = self.efficiency or agent.efficiency
        duty = duty/efficiency
        if agent.Hvap:
            # Phase change
            self.outlet_utility_stream.phase = 'g' if self.inlet_utility_stream.phase == 'l' else 'l'
            dH = agent.Hvap * self.outlet_utility_stream.F_mol
        else:
            # Temperature change
            self.outlet_utility_stream.T = self._T_exit(T_pinch_out, agent.T_limit, iscooling)
            dH = self.inlet_utility_stream.H - self.outlet_utility_stream.H
        
        # Update utility flow
        self.outlet_utility_stream.mol[:] *= duty / dH
        
        # Update and return results
        self.flow = F_mol = self.inlet_utility_stream.F_mol
        self.duty = duty
        self.cost = agent.price_kJ*duty + agent.price_kmol*F_mol

    @staticmethod
    def _get_pinch(iscooling, T_in, T_out, dT):
        """Return pinch entrance and exit temperature of utility."""
        if iscooling:
            assert T_in + 1e-6 >= T_out, "entrance temperature must be larger than exit temperature if cooling"
            T_pinch_in = T_out - dT
            T_pinch_out = T_in - dT
        else:
            assert T_in <= T_out + 1e-6, "entrance temperature must be smaller than exit temperature if heating"
            T_pinch_in = T_out + dT
            T_pinch_out = T_in + dT
        return T_pinch_in, T_pinch_out
    
    def load_agent(self, ID, agent):
        if self.ID != ID:
            self._init_streams(agent.z_mol, agent.thermo,
                               agent.T, agent.P, agent.phase)
            self.ID = ID

    # Subcalculations

    @staticmethod
    def _T_exit(T_pinch, T_limit, iscooling):
        """Return exit temperature of the utility in a counter current heat exchanger

        Parameters
        ----------
        T_pinch : float
                  Pinch temperature of utility stream (K)
        iscooling : bool
                  True if exit temperature should be lower (process stream is gaining energy)

        """
        if iscooling:
            return T_limit if T_limit and T_limit < T_pinch else T_pinch
        else:
            return T_limit if T_limit and T_limit > T_pinch else T_pinch
        

    def _info_data(self, duty, flow, cost):
        # Get units of measure
        su = self.display_units
        duty_units = duty or su.duty
        flow_units = flow or su.flow
        cost_units = cost or su.cost
        
        # Change units and return info string
        flow = self.inlet_utility_stream.get_total_flow(flow_units)
        duty = convert(self.duty, 'kJ/hr', duty_units)
        cost = convert(self.cost, 'USD/hr', cost_units)
        return duty, flow, cost, duty_units, flow_units, cost_units
        
    def __repr__(self):
        if self.ID:
            duty, flow, cost, duty_units, flow_units, cost_units = self._info_data(None, None, None)
            return f'<{self.ID}: {self.duty:.3g} {duty_units}, {self.flow:.3g} {flow_units}, {self.cost:.3g} {cost_units}>'
        else:
            return f'<{type(self).__name__}: None>'
        
    # Representation
    def _info(self, duty, flow, cost):
        """Return string related to specifications"""
        if not self.ID:
            return (f'{type(self).__name__}: None\n'
                    +f' duty: 0\n'
                    +f' flow: 0\n'
                    +f' cost: 0')
        else:
            (duty, flow, cost, duty_units,
             flow_units, cost_units) = self._info_data(duty, flow, cost)
            return (f'{type(self).__name__}: {self.ID}\n'
                    +f' duty:{duty: .3g} {duty_units}\n'
                    +f' flow:{flow: .3g} {flow_units}\n'
                    +f' cost:{cost: .3g} {cost_units}')
            

    def show(self, duty=None, flow=None, cost=None):
        """Print all specifications"""
        print(self._info(duty, flow, cost))
    _ipython_display_ = show


del _Water


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 16:42:49 2018

This module includes classes and functions relating arrays.

@author: Guest Group
"""
import numpy as np

__all__ = ('fraction',)

ndarray = np.ndarray
asarray = np.asarray

# %% Functions

def fraction(array):
    """Return a normalized array to a magnitude of 1. If magnitude is zero, all fractions will have equal value."""
    sum_array = array.sum()
    if sum_array == 0:
        l = len(array)
        return np.ones(l)/l
    else:
        return array/sum_array
    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 16:21:56 2018

@author: yoelr
"""
from .. import Unit, PowerUtility
from thermosteam import MultiStream
from math import pi, ceil
import numpy as np
from .designtools import calculate_vacuum_system_power_and_cost, HNATable, FinalValue, \
                          VesselWeightAndWallThickness, Kvalue
from ._splitter import Splitter
from ._hx import HX, HXutility
from ..utils import bounds_warning
import biosteam as bst
from ._CAS import H2O_CAS

exp = np.exp
ln = np.log

# USDA Biodiesel: 
#V = np.pi*D**2/4*Ht*0.3048**3
#if not (0.1 < V < 70):
#    raise DesignError(f"Volume is out of bounds for costing")
#lambda V, CE: CE*13*V**0.62 # V (m3)
__all__ = ('Flash', 'SplitFlash', 'RatioFlash')


# %% Data
    
# Material density (lb∕ft^3)
material_density = {'Carbon steel': 490,
                    'Low-alloy steel': None,
                    'Stainless steel 304': 499.4,
                    'Stainless steel 316': 499.4,
                    'Carpenter 20CB-3': None,
                    'Nickel-200': None,
                    'Monel-400': None,
                    'Inconel-600': None,
                    'Incoloy-825': None,
                    'Titanium': None}

# Vessel Material
material_factor = {'Carbon steel': 1.0,
                   'Low-alloy steel': 1.2,
                   'Stainless steel 304': 1.7,
                   'Stainless steel 316': 2.1,
                   'Carpenter 20CB-3': 3.2,
                   'Nickel-200': 5.4,
                   'Monel-400': 3.6,
                   'Inconel-600': 3.9,
                   'Incoloy-825': 3.7,
                   'Titanium': 7.7}


# %% Flash

class Flash(Unit):
    """Create an equlibrium based flash drum with the option of having light non-keys and heavy non-keys completly separate into their respective phases. Design procedure is based on heuristics by Wayne D. Monnery & William Y. Svrcek [1]. Purchase costs are based on correlations by Mulet et al. [2, 3] as compiled by Warren et. al. [4].

    Parameters
    ----------
    ins
        [0] Input stream
        
    outs
        [0] Vapor product
        
        [1] Liquid product
    Specify two:
        * **P:** Operating pressure (Pa)
        * **Q:** Energy input (kJ/hr)
        * **T:** Operating temperature (K)
        * **V:** Molar vapor fraction
        * **x:** Molar composition of liquid (for binary mixture)
        * **y:** Molar composition of vapor (for binary mixture)

    References
    ----------
    [1] "Design Two-Phase Separators Within the Right Limits", Chemical Engineering Progress Oct, 1993.

    [2] Mulet, A., A. B. Corripio, and L. B. Evans, “Estimate Costs of Pressure Vessels via Correlations,” Chem. Eng., 88(20), 145–150 (1981a).

    [3] Mulet, A., A.B. Corripio, and L.B.Evans, “Estimate Costs of Distillation and Absorption Towers via Correlations,” Chem. Eng., 88(26), 77–82 (1981b).

    [4] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley. Cost Accounting and Capital Cost Estimation (Chapter 16)    
    
    Examples
    --------
    :doc:`notebooks/Flash Example`
    
    """
    _units = {'Vertical vessel weight': 'lb',
              'Horizontal vessel weight': 'lb',
              'Length': 'ft',
              'Diameter': 'ft',
              'Weight': 'lb',
              'Wall thickness': 'in'}
    _has_power_utility = False
    _N_heat_utilities = 0
    
    # Bare module factor
    BM_horizontal = 3.05
    BM_vertical = 4.16
    @property
    def BM(self):
        SepType = self.SetType
        if SepType == 'Vertical':
            return self._BM_vertical
        elif SepType == 'Horizontal':
            return self._BM_horizontal
        else:
            raise AttributeError('SepType not defined')
    
    # Column material factor
    _material = 'Carbon steel'
    _F_material = 1 
    
    #: If a vacuum system is needed, it will choose one according to this preference.
    vacuum_system_preference = 'Liquid-ring pump'
    
    #: True if glycol groups are present in the mixture
    HasGlycolGroups = False
    
    #: True if amine groups are present in the mixture
    HasAmineGroups = False
    
    #: [str] 'Horizontal', 'Vertical', or 'Default'
    SepType = 'Default'
    
    #: [float] Time it takes to raise liquid to half full (min)
    HoldupTime = 15  
    
    #: [float] Time it takes to reach from normal to maximum liquied level (min)
    SurgeTime = 7.5
    
    #: [bool] True if using a mist eliminator pad
    Mist = False
    
    _bounds = {'Vertical vessel weight': (4200, 1e6),
               'Horizontal vessel weight': (1e3, 9.2e5),
               'Diameter': (3, 21),
               'Vertical vessel length': (12, 40)}

    @property
    def material(self):
        """Vessel construction material."""
        return self._material
    @material.setter
    def material(self, material):
        try: self._F_material = material_factor[material]
        except KeyError:
            dummy = str(material_factor.keys())[11:-2]
            raise ValueError(f"material must be one of the following: {dummy}")
        self._material = material  

    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 keys=None, LNK=(), HNK=(),
                 V=None, T=None, Q=None, P=None,
                 y=None, x=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        self._multistream = MultiStream(None)
        self._heat_exchanger = None
        
        #: tuple[str] IDs of chemicals in thermodynamic equilibrium
        self.keys = keys
        
        #: tuple[str] Light non-keys assumed to remain as a vapor
        self.LNK = LNK
        
        #: tuple[str] Heavy non-keys assumed to remain as a liquid
        self.HNK = HNK
        
        #: Enforced molar vapor fraction
        self.V = V
        
        #: Enforced operating temperature (K)
        self.T = T
        
        #: [array_like] Molar composition of vapor (for binary mixture)
        self.y = y
        
        #: [array_like] Molar composition of liquid (for binary mixture)
        self.x = x
        
        #: Duty (kJ/hr)
        self.Q = Q
        
        #: Operating pressure (Pa)
        self.P = P
        
    @property
    def P(self):
        """Operating pressure (Pa)."""
        return self._P
    @P.setter
    def P(self, P):
        if P < 101325 and not self.power_utility:
            self.power_utility = PowerUtility()
        self._P = P
    
    @property
    def Q(self):
        """Enforced duty (kJ/hr)."""
        return self._Q
    @Q.setter
    def Q(self, Q):
        if Q == 0:
            self._heat_exchanger = None
        elif not self._heat_exchanger:
            self._heat_exchanger = he = HXutility(None, outs=None) 
            self.heat_utilities = he.heat_utilities
            he._ins = self._ins
            he._outs[0] = self._multistream
        self._Q = Q

    def _run(self):
        # Unpack
        vap, liq = self.outs
        feed = self.ins[0]

        # Vapor Liquid Equilibrium
        ms = self._multistream
        ms.empty()
        ms.imol['l'] = feed.mol
        ms.T = feed.T
        Q = self.Q
        H = feed.H + Q if Q is not None else None
        ms.vle(self.P, H, self.T, self.V, self.x, self.y)

        # Set Values
        vap.phase = 'g'
        liq.phase = 'l'
        vap.mol[:] = ms.imol['g']
        liq.mol[:] = ms.imol['l']
        vap.T = liq.T = ms.T
        vap.P = liq.P = ms.P

    def _design(self):
        # Set horizontal or vertical vessel
        SepType = self.SepType
        if SepType == 'Default':
            vap, liq = self.outs
            isVertical = vap.F_mass/liq.F_mass > 0.2
        elif SepType == 'Vertical':
            isVertical = True
        elif SepType == 'Horizontal':
            isVertical = False
        else:
            raise ValueError( f"SepType must be either 'Default', 'Horizontal', 'Vertical', not '{self.SepType}'")
        self._isVertical = isVertical

        # Run vertical or horizontal design
        if isVertical: self._vertical()
        else: self._horizontal()
        if self._heat_exchanger: self._heat_exchanger._design()
        self.design_results['Material'] = self._material

    def _cost(self):
        Design = self.design_results
        W = Design['Weight']
        D = Design['Diameter']
        L = Design['Length']
        CE = bst.CE
        
        # C_v: Vessel cost
        # C_pl: Platforms and ladders cost
        if self._isVertical:
            C_v = exp(7.1390 + 0.18255*ln(W) + 0.02297*ln(W)**2)
            C_pl = 410*D**0.7396*L**0.70684
        else:
            C_v = exp(5.6336 - 0.4599*ln(W) + 0.00582*ln(W)**2)
            C_pl = 2275*D**0.20294
            
        self.purchase_costs['Flash'] = CE/567*(self._F_material*C_v+C_pl)
        if self._heat_exchanger:
            hx = self._heat_exchanger
            hx._cost()
            self.purchase_costs.update(hx.purchase_costs)
        self._cost_vacuum()

    def _cost_vacuum(self):
        P = self.P
        if not P or P > 101320: return 
        
        Design = self.design_results
        volume = 0.02832 * np.pi * Design['Length'] * (Design['Diameter']/2)**2
        
        # If vapor is condensed, assume vacuum system is after condenser
        vapor = self.outs[0]
        hx = vapor.sink
        if isinstance(hx, HX):
            index = hx.ins.index(vapor)
            stream = hx.outs[index]
            if isinstance(stream, MultiStream):
                vapor = stream['g']
                F_mass = vapor.F_mass
                F_vol = vapor.F_vol
            else:
                if stream.phase == 'g':
                    F_mass = stream.F_mass
                    F_vol = stream.F_vol
                else:
                    F_mass = 0
                    F_vol = 0
        else:
            F_mass = vapor.F_mass
            F_vol = vapor.F_vol
        
        power, cost = calculate_vacuum_system_power_and_cost(
                          F_mass, F_vol, P, volume, self.vacuum_system_preference)
        self.purchase_costs['Liquid-ring pump'] = cost
        self.power_utility(power)

    def _design_parameters(self):
        # Retrieve run_args and properties
        vap, liq = self._outs
        rhov = vap.get_property('rho', 'lb/ft3')
        rhol = liq.get_property('rho', 'lb/ft3')
        P = liq.P*0.000145  # Pressure (psi)

        # SepType (str), HoldupTime (min), SurgeTime (min), Mist (bool)
        SepType = self.SepType
        Th = self.HoldupTime
        Ts = self.SurgeTime
        Mist = self.Mist

        # Calculate the volumetric flowrate
        Qv = vap.get_total_flow('ft^3 / s')
        Qll = liq.get_total_flow('ft^3 / min')

        # Calculate Ut and set Uv
        K = Kvalue(P)

        # Adjust K value
        if not Mist and SepType == 'Vertical': K /= 2

        # Adjust for amine or glycol groups:
        if self.HasGlycolGroups: K *= 0.6
        elif self.HasAmineGroups: K *= 0.8

        Ut = K*((rhol - rhov) / rhov)**0.5
        Uv = 0.75*Ut

        # Calculate Holdup and Surge volume
        Vh = Th*Qll
        Vs = Ts*Qll
        return rhov, rhol, P, Th, Ts, Mist, Qv, Qll, Ut, Uv, Vh, Vs

    def _vertical(self):
        rhov, rhol, P, Th, Ts, Mist, Qv, Qll, Ut, Uv, Vh, Vs = self._design_parameters()

        # Calculate internal diameter, Dvd
        Dvd = (4.0*Qv/(pi*Uv))**0.5
        if Mist:
            D = FinalValue(Dvd + 0.4)
        else:
            D = FinalValue(Dvd)

        # Obtaining low liquid level height, Hlll
        Hlll = 0.5
        if P < 300:
            Hlll = 1.25

        # Calculate the height from Hlll to  Normal liq level, Hnll
        Hh = Vh/(pi/4.0*Dvd**2)
        Hh = 1.0 if Hh < 1.0 else FinalValue(Hh)

        # Calculate the height from Hnll to  High liq level, Hhll
        Hs = Vs/(pi/4.0*Dvd**2)
        Hs = 0.5 if Hs < 0.5 else FinalValue(Hs)

        # Calculate dN
        Qm = Qll + Qv
        lamda = Qll/Qm
        rhoM = rhol*lamda + rhov*(1-lamda)
        dN = (4*Qm/(pi*60.0/(rhoM**0.5)))**0.5
        dN = FinalValue(dN)

        # Calculate Hlin, assume with inlet diverter
        Hlin = 1.0 + dN

        # Calculate the vapor disengagement height
        Hv = 0.5*Dvd
        Hv2 = (2.0 if Mist else 3.0) + dN/2.0
        if Hv2 < Hv: Hv = Hv2
        Hv = ceil(Hv)

        # Calculate total height, Ht
        Hme = 1.5 if Mist else 0.0
        Ht = Hlll + Hh + Hs + Hlin + Hv + Hme
        Ht = FinalValue(Ht)

        # Check if LD is between 1.5 and 6.0
        while True:
            LD = Ht/D
            if (LD < 1.5): D -= 0.5
            elif (LD > 6.0): D += 0.5
            else: break

        # Calculate Vessel weight and wall thickness
        rho_M = material_density[self._material]
        VW, VWT = VesselWeightAndWallThickness(P, D, Ht, rho_M)

        # Find maximum and normal liquid level
        # Hhll = Hs + Hh + Hlll
        # Hnll = Hh + Hlll

        Design = self.design_results
        bounds_warning(self, 'Vertical vessel weight', VW, 'lb', self._bounds['Vertical vessel weight'])
        bounds_warning(self, 'Vertical vessel length', Ht, 'ft', self._bounds['Vertical vessel length'])
        Design['SepType'] = 'Vertical'
        Design['Length'] = Ht     # ft
        Design['Diameter'] = D    # ft
        Design['Weight'] = VW     # lb
        Design['Wall thickness'] = VWT  # in
        
    def _horizontal(self):
        rhov, rhol, P, Th, Ts, Mist, Qv, Qll, Ut, Uv, Vh, Vs = self._design_parameters()

        # Initialize LD
        if P > 0 and P <= 264.7:
            LD = 1.5/250.0*(P-14.7)+1.5
        elif P > 264.7 and P <= 514.7:
            LD = 1.0/250.0*(P-14.7)+2.0
        elif P > 514.7:
            LD = 5.0

        D = (4.0*(Vh+Vs)/(0.6*pi*LD))**(1.0/3.0)
        if D <= 4.0:
            D = 4.0
        else:
            D = FinalValue(D)

        for outerIter in range(50):
            At = pi*(D**2)/4.0 # Total area

            # Calculate Lower Liquid Area
            Hlll = round(0.5*D + 7.0)  
            Hlll = Hlll/12.0 # D is in ft but Hlll is in inches
            X = Hlll/D
            Y = HNATable(1, X)
            Alll = Y*At

            # Calculate the Vapor disengagement area, Av
            Hv = 0.2*D
            if Mist and Hv <= 2.0: Hv = 2.0
            elif Hv <= 1.0: Hv = 1.0
            else: Hv = FinalValue(Hv)
            Av = HNATable(1, Hv/D)*At
            
            # Calculate minimum length for surge and holdup
            L = (Vh + Vs)/(At - Av - Alll)
            # Calculate liquid dropout
            Phi = Hv/Uv
            # Calculate actual vapor velocity
            Uva = Qv/Av
            # Calculate minimum length for vapor disengagement
            Lmin = Uva*Phi
            Li = L
            
            for innerIter in range(50):
                if L < 0.8*Lmin: Hv += 0.5
                elif L > 1.2*Lmin:
                    if Mist and Hv <= 2.0: Hv = 2.0
                    elif not Mist and Hv <= 1.0: Hv = 1.0
                    else: Hv -= 0.5
                else: break
                Av = HNATable(1, Hv/D)*At
                Alll = HNATable(1, Hlll/D)*At
                Li = (Vh + Vs)/(At - Av - Alll)
                Phi = Hv/Uv
                Uva = Qv/Av
                Lmin = Uva*Phi
            
            L = Li
            LD = L/D
            # Check LD
            if LD < 1.2:
                if D <= 4.0: break
                else: D -= 0.5

            if LD > 7.2:
                D += 0.5
            else: break

        # Recalculate LD so it lies between 1.5 - 6.0
        while True:
            LD = L / D
            if (LD < 1.5) and D <= 4.0: L += 0.5
            elif LD < 1.5: D -= 0.5
            elif (LD > 6.0): D += 0.5
            else: break

        # Calculate vessel weight and wall thickness
        rho_M = material_density[self._material]
        VW, VWT = VesselWeightAndWallThickness(P, D, L, rho_M)

        # # To check minimum Hv value
        # if int(Mist) == 1 and Hv <= 2.0:
        #     Hv = 2.0
        # if int(Mist) == 0 and Hv <= 1.0:
        #     Hv = 1.0

        # Calculate normal liquid level and High liquid level
        # Hhll = D - Hv
        # if (Hhll < 0.0):
        #     Hhll = 0.0
        # Anll = Alll + Vh/L
        # X = Anll/At
        # Y = HNATable(2, X)
        # Hnll = Y*D
        
        Design = self.design_results
        bounds_warning(self, 'Horizontal vessel weight', VW, 'lb', self._bounds['Horizontal vessel weight'])
        Design['SepType'] = 'Horizontal'
        Design['Length'] = L  # ft
        Design['Diameter'] = D  # ft
        Design['Weight'] = VW  # lb
        Design['Wall thickness'] = VWT  # in    
        
    def _end_decorated_cost_(self):
        if self.heat_utilities: self.heat_utilities[0](self.Hnet, self.T)
    

# %% Special

class SplitFlash(Flash):
    line = 'Flash' 
    
    def __init__(self, ID='', ins=None, outs=(), *, split,
                 order=None, T=None, P=None, Q=None):
        Splitter.__init__(self, ID, ins, outs, split=split, order=order)
        self._multistream = MultiStream(None)
        self._heat_exchanger = None
        self.T = T #: Operating temperature (K)
        self.P = P #: Operating pressure (Pa)
        self.Q = Q #: Duty (kJ/hr)
    
    split = Splitter.split
    V = None
    
    def _run(self):
        top, bot = self.outs
        feed, = self.ins
        feed_mol = feed.mol
        top.mol[:] = top_mol = feed_mol * self.split
        bot.mol[:] = feed_mol - top_mol
        top.phase = 'g'
        bot.phase = 'l'
        bot.T = top.T = self.T
        bot.P = top.P = self.P

    def _design(self):
        if self._heat_exchanger:
            self._heat_exchanger.outs[0] = ms = self._multistream
            ms.mix_from(self.outs)
        super()._design()


# class PartitionFlash(Flash):
#     """Create a PartitionFlash Unit that approximates outputs based on molar partition coefficients."""
    
#     _kwargs = {'species_IDs': None,  # Partition species
#                'Ks': None,           # Partition coefficients
#                'LNK': None,          # light non-keys
#                'HNK': None,          # heavy non-keys
#                'P': None,            # Operating Pressure
#                'T': None}            # Operating temperature
    
#     def _init(self):
#         top, bot = self.outs
#         species_IDs, Ks, LNK, HNK, P, T = self._kwargs.values()
#         index = top._species._IDs.index
#         if P < 101325 and not self.power_utility:
#             self.power_utility = PowerUtility()
#         self._args = ([index(i) for i in species_IDs],
#                       [index(i) for i in LNK],
#                       [index(i) for i in HNK],
#                        np.asarray(Ks), P, T)
    
#     def _run(self):
#         feed = self.ins[0]
#         ph1, ph2 = self.outs
#         pos, LNK, HNK, Ks, P, T = self._args
#         mol_in = feed.mol[pos]
#         ph1.T = ph2.T = (T if T else feed.T)
#         ph1.P = ph2.P = (P if P else feed.P)
#         ph1.mol[LNK] = feed.mol[LNK]
#         ph1.mol[HNK] = feed.mol[HNK]
        
#         # Get zs and Ks to solve rashford rice equation
#         molnet = sum(mol_in)
#         zs = mol_in/molnet
#         lenzs = len(zs)
#         if hasattr(self, '_V'):
#             self._V = V = newton(V_error, self._V, args=(zs, Ks))
#         elif lenzs > 3:
#             self._V = V = brentq(V_error, 0, 1, (zs, Ks))
#         elif lenzs == 2:
#             self._V = V = V_2N(V, zs, Ks)
#         elif lenzs == 3:
#             self._V = V = V_3N(V, zs, Ks)
            
#         x = zs/(1 + V*(Ks-1))
#         y = x*Ks
#         ph1.mol[pos] = molnet*V*y
#         ph2.mol[pos] = mol_in - ph1.mol[pos]
    
#     _design = SplitFlash._design
    

class RatioFlash(Flash):
    _N_heat_utilities = 1

    def __init__(self, ID='', ins=None, outs=(), *,
                 K_chemicals, Ks, top_solvents=(), top_split=(),
                 bot_solvents=(), bot_split=()):
        Unit.__init__(self, ID, ins, outs)
        self.K_chemicals = K_chemicals
        self.Ks = Ks
        self.top_solvents = top_solvents
        self.top_split = top_split
        self.bot_solvents = bot_solvents
        self.bot_split = bot_split

    def _run(self):
        feed = self.ins[0]
        top, bot = self.outs
        indices = self.chemicals.get_index
        K_index = indices(self.K_chemicals)
        top_index = indices(self.top_solvents)
        bot_index = indices(self.bot_solvents)
        top_mol = top.mol; bot_mol = bot.mol; feed_mol = feed.mol
        top_mol[top_index] = feed_mol[top_index]*self.top_split
        bot_mol[top_index] = feed_mol[top_index]-top_mol[top_index]
        bot_mol[bot_index] = feed_mol[bot_index]*self.bot_split
        top_mol[bot_index] = feed_mol[bot_index]-bot_mol[bot_index]
        topnet = top_mol[top_index].sum()
        botnet = bot_mol[bot_index].sum()
        molnet = topnet+botnet
        top_mol[K_index] = self.Ks * topnet * feed_mol[K_index] / molnet  # solvent * mol ratio
        bot_mol[K_index] = feed_mol[K_index] - top_mol[K_index]
        top.T, top.P = feed.T, feed.P
        bot.T, bot.P = feed.T, feed.P


# %% Single Component

class Evaporator_PQ(Unit):
    _N_ins = 2
    _N_outs = 3
    @property
    def P(self):
        return self._P
    @P.setter
    def P(self, P):
        water = getattr(self.chemicals, H2O_CAS)
        self._T = T = water.Tsat(P)
        self._Hvap = water.Hvap(T)
        self._P = P
    @property
    def T(self):
        return self._T
    @T.setter
    def T(self, T):
        water = getattr(self.chemicals, H2O_CAS)
        self._P = water.Psat(T)
        self._Hvap = water.Hvap(T)
        self._T = T
    
    def __init__(self, ID='', ins=None, outs=(), *, Q=0, P=101325):
        super().__init__(ID, ins, outs)
        self.Q = Q
        self.P = P
    
    def _run(self):
        feed, utility_vapor = self.ins
        vapor, liquid, utility_liquid = self.outs
        
        # Optional if Q also comes from condensing a side stream
        Q = self.Q
        if utility_liquid:
            utility_liquid.copy_like(utility_vapor)
            utility_liquid.phase = 'l'
            Q += utility_vapor.H - utility_liquid.H
        feed_H = feed.H
    
        # Set exit conditions
        vapor.T = liquid.T = self.T
        vapor.P = liquid.P = self.P
        liquid.phase = 'l'
        vapor.phase = 'g'
        liquid.copy_flow(feed, IDs=H2O_CAS, exclude=True)
        
        # Energy balance to find vapor fraction
        f = feed.imol[H2O_CAS]
        H = feed_H + Q - liquid.H
        V = (H/f)/(self._Hvap)
        if V < 0:
            V = 0
        elif V > 1:
            V = 1
        evaporated = f * V
        vapor.imol[H2O_CAS] = evaporated
        liquid.imol[H2O_CAS] = (1-V)*f
        self._Q = Q
        self._V = V


class Evaporator_PV(Unit):
    _N_heat_utilities = 1

    @property
    def P(self):
        return self._P
    @P.setter
    def P(self, P):
        water = getattr(self.chemicals, H2O_CAS)
        self._T = water.Tsat(P)
        self._P = P
    @property
    def T(self):
        return self._T
    @T.setter
    def T(self, T):
        water = getattr(self.chemicals, H2O_CAS)
        self._P = water.Psat(T)
        self._T = T
    
    @property
    def component(self):
        return self._component

    def __init__(self, ID='', ins=None, outs=(), *, V=0.5, P=101325):
        super().__init__(ID, ins, outs)
        self.V = V
        self.P = P

    def _run(self):
        feed = self.ins[0]
        vapor, liquid = self.outs
        H2O_index = self.chemicals.index(H2O_CAS)
        water_mol = feed.mol[H2O_index]
        vapor.mol[H2O_index] = self.V * water_mol
        liquid_mol = liquid.mol
        liquid_mol[:] = feed.mol
        liquid_mol[H2O_index] = (1-self.V) * water_mol

        